/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif

#include "UtlException.h"
#include "ScpStream.h"
#include "CppCompilerEnvironment.h"
#include "CGRuntime.h"

#include "DtaScriptVariable.h"
#include "ExprScriptVariable.h"
#include "DtaBNFScript.h"
#include "BNFClause.h"
#include "BNFCharBoundaries.h"

using namespace CodeWorker;

BNFCharBoundaries::BNFCharBoundaries(DtaBNFScript* pBNFScript, GrfBlock* pParent, int iStartChar, int iEndChar, bool bContinue, bool bNoCase) : GrfCommand(pParent), _pBNFScript(pBNFScript), _iStartChar(iStartChar), _iEndChar(iEndChar), _pVariableToAssign(0L), _bContinue(bContinue), _iClauseReturnType(BNFClause::NO_RETURN_TYPE), _bNoCase(bNoCase) {
	if (_bNoCase) {
		if ((iStartChar >= (int) 'A') && (_iStartChar <= (int) 'Z')) iStartChar += 32;
		if ((iEndChar >= (int) 'A') && (_iEndChar <= (int) 'Z')) iEndChar += 32;
		if ((iStartChar >= (int) 'a') && (iStartChar <= (int) 'z') &&
			(iEndChar >= (int) 'a') && (iEndChar <= (int) 'z')) {
			_iStartChar = iStartChar;
			_iEndChar = iEndChar;
		} else {
			_bNoCase = false;
		}
	}
}

BNFCharBoundaries::~BNFCharBoundaries() {
	delete _pVariableToAssign;
}

bool BNFCharBoundaries::isABNFCommand() const { return true; }

void BNFCharBoundaries::setVariableToAssign(ExprScriptVariable* pVariableToAssign, bool bConcat, BNFClause& theClause) {
	if (pVariableToAssign != 0L) {
		_pVariableToAssign = pVariableToAssign;
		std::string sVariableName = _pVariableToAssign->toString();
		if (sVariableName == theClause.getName()) _iClauseReturnType = theClause.getReturnType();
		_bConcatVariable = bConcat;
	}
}

SEQUENCE_INTERRUPTION_LIST BNFCharBoundaries::executeInternal(DtaScriptVariable& visibility) {
	int iLocation = CGRuntime::getLocation();
	int iImplicitCopyPosition = _pBNFScript->skipEmptyChars(visibility);
	int iChar = CGRuntime::readCharAsInt();
	if (_bNoCase && (iChar <= (int) 'Z')) iChar += 32;
	if ((iChar < 0) || (iChar < _iStartChar) || (iChar > _iEndChar)) {
		if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
		CGRuntime::setLocation(iLocation);
		if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
		return BREAK_INTERRUPTION;
	}
	if (!_listOfConstants.empty()) {
		bool bMatch = false;
		for (std::vector<std::string>::const_iterator i = _listOfConstants.begin(); i != _listOfConstants.end(); i++) {
			bMatch = ((int) ((*i)[0]) == iChar);
			if (bMatch) break;
		}
		if (!bMatch) {
			if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
			CGRuntime::setLocation(iLocation);
			if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
			return BREAK_INTERRUPTION;
		}
	}
	if (_pVariableToAssign != 0L) {
		DtaScriptVariable* pVariable = visibility.getVariable(*_pVariableToAssign);
		char tcText[] = {'\0', '\0'};
		tcText[0] = (char) iChar;
		if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
			pVariable->pushItem(tcText);
		} else {
			if (_bConcatVariable) pVariable->concatValue(tcText);
			else pVariable->setValue(tcText);
		}
	}
	if (iImplicitCopyPosition >= 0) {
		register char cChar = (char) iChar;
		_pBNFScript->writeBinaryData(&cChar, 1);
	}
	return NO_INTERRUPTION;
}

void BNFCharBoundaries::compileCpp(CppCompilerEnvironment& theCompilerEnvironment) const {
	int iCursor = theCompilerEnvironment.newCursor();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "// " << toString();
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseLocation_" << iCursor << " = CGRuntime::getLocation();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseImplicitCopy_" << iCursor << " = theEnvironment.skipEmptyChars();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "char _compilerClauseChar_" << iCursor << " = CGRuntime::readChar()[0];";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseChar_" << iCursor << " == '\\0') {";
	theCompilerEnvironment.getBody().endl();
	if (_bContinue) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::throwBNFExecutionError(";
		theCompilerEnvironment.getBody().writeString(toString());
		theCompilerEnvironment.getBody() << ");";
	} else {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t_compilerClauseSuccess = false;";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::setLocation(_compilerClauseLocation_" << iCursor << ")";
	}
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "} else {";
	theCompilerEnvironment.getBody().endl();
	if (_bNoCase) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	if (_compilerClauseChar_" << iCursor << " <= (int) 'Z') _compilerClauseChar_" << iCursor << " += 32;";
		theCompilerEnvironment.getBody().endl();
	}
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	_compilerClauseSuccess = (_compilerClauseChar_" << iCursor << " >= ";
	theCompilerEnvironment.getBody().writeQuotedChar(_iStartChar);
	theCompilerEnvironment.getBody() << ") && (_compilerClauseChar_" << iCursor << " <= ";
	theCompilerEnvironment.getBody().writeQuotedChar(_iEndChar);
	theCompilerEnvironment.getBody() << ");";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	if (!_compilerClauseSuccess) ";
	if (_bContinue) {
		theCompilerEnvironment.getBody() << "CGRuntime::throwBNFExecutionError(";
		theCompilerEnvironment.getBody().writeString(toString());
		theCompilerEnvironment.getBody() << ");";
	} else {
		theCompilerEnvironment.getBody() << "CGRuntime::setLocation(_compilerClauseLocation_" << iCursor << ");";
	}
	theCompilerEnvironment.getBody().endl();
	if (!_listOfConstants.empty()) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tswitch(_compilerClauseChar_" << iCursor << ") {";
		theCompilerEnvironment.getBody().endl();
		for (std::vector<std::string>::const_iterator i = _listOfConstants.begin(); i != _listOfConstants.end(); i++) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t\tcase " << (int) ((*i)[0]) << ":";
			theCompilerEnvironment.getBody().endl();
		}
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t\t\tbreak;";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t\tdefault: ";
		if (_bContinue) {
			theCompilerEnvironment.getBody() << "CGRuntime::throwBNFExecutionError(";
			theCompilerEnvironment.getBody().writeString(toString());
			theCompilerEnvironment.getBody() << ");";
		} else {
			theCompilerEnvironment.getBody() << "_compilerClauseSuccess = false;CGRuntime::setLocation(_compilerClauseLocation_" << iCursor << ");";
		}
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t}";
		theCompilerEnvironment.getBody().endl();
	}
	if (_pVariableToAssign != 0L) {
		if (!_listOfConstants.empty() && !_bContinue) {
			theCompilerEnvironment.incrementIndentation();
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
			theCompilerEnvironment.getBody().endl();
		}
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tchar tcText[] = {'\\0', '\\0'};";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\ttcText[0] = (char) _compilerClauseChar_" << iCursor << ";";
		theCompilerEnvironment.getBody().endl();
		if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".pushItem(tcText);";
			theCompilerEnvironment.getBody().endl();
		} else {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".setValue(tcText);";
			theCompilerEnvironment.getBody().endl();
		}
		if (!_listOfConstants.empty() && !_bContinue) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
			theCompilerEnvironment.getBody().endl();
			theCompilerEnvironment.decrementIndentation();
		}
	}
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
	theCompilerEnvironment.getBody().endl();
}

std::string BNFCharBoundaries::toString() const {
	std::string sString;
	if (_bContinue) sString = "#continue ";
	sString += "'";
	char a = (char) _iStartChar;
	switch(a) {
		case '\\': sString += "\\\\";break;
		case '\"': sString += "\\\"";break;
		case '\'': sString += "\\'";break;
		case '\a': sString += "\\a";break;
		case '\b': sString += "\\b";break;
		case '\f': sString += "\\f";break;
		case '\n': sString += "\\n";break;
		case '\r': sString += "\\r";break;
		case '\t': sString += "\\t";break;
		case '\v': sString += "\\v";break;
		default:
			sString += a;
	}
	sString += "'..'";
	a = (char) _iEndChar;
	switch(a) {
		case '\\': sString += "\\\\";break;
		case '\"': sString += "\\\"";break;
		case '\'': sString += "\\'";break;
		case '\a': sString += "\\a";break;
		case '\b': sString += "\\b";break;
		case '\f': sString += "\\f";break;
		case '\n': sString += "\\n";break;
		case '\r': sString += "\\r";break;
		case '\t': sString += "\\t";break;
		case '\v': sString += "\\v";break;
		default:
			sString += a;
	}
	sString += "'";
	return sString + DtaBNFScript::constantsToString(_listOfConstants) + DtaBNFScript::assignmentToString(_pVariableToAssign);
}
