/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif

#include "UtlException.h"
#include "ScpStream.h"
#include "CppCompilerEnvironment.h"
#include "CGRuntime.h"

#include "DtaScriptVariable.h"
#include "ExprScriptVariable.h"
#include "DtaBNFScript.h"
#include "BNFClause.h"
#include "BNFCharLitteral.h"

using namespace CodeWorker;

BNFCharLitteral::BNFCharLitteral(DtaBNFScript* pBNFScript, GrfBlock* pParent, int iChar, bool bContinue, bool bNoCase) : _pBNFScript(pBNFScript), GrfCommand(pParent), _iChar(iChar), _pVariableToAssign(0L), _bContinue(bContinue), _iClauseReturnType(BNFClause::NO_RETURN_TYPE), _bNoCase(bNoCase) {
	if (_bNoCase) {
		if ((_iChar >= (int) 'A') && (_iChar <= (int) 'Z')) iChar += 32;
		if ((iChar >= (int) 'a') && (iChar <= (int) 'z')) _iChar = iChar;
		else _bNoCase = false;
	}
}

BNFCharLitteral::~BNFCharLitteral() {
	delete _pVariableToAssign;
}

bool BNFCharLitteral::isABNFCommand() const { return true; }

void BNFCharLitteral::setVariableToAssign(ExprScriptVariable* pVariableToAssign, bool bConcat, BNFClause& theClause) {
	if (pVariableToAssign != 0L) {
		_pVariableToAssign = pVariableToAssign;
		std::string sVariableName = _pVariableToAssign->toString();
		if (sVariableName == theClause.getName()) _iClauseReturnType = theClause.getReturnType();
		_bConcatVariable = bConcat;
	}
}

SEQUENCE_INTERRUPTION_LIST BNFCharLitteral::executeInternal(DtaScriptVariable& visibility) {
	int iLocation = CGRuntime::getLocation();
	int iImplicitCopyPosition = _pBNFScript->skipEmptyChars(visibility);
	int iChar = CGRuntime::readCharAsInt();
	if (iChar != _iChar) {
		if (_bNoCase && (iChar + 32 == _iChar)) {
			// OK! Difference was due to upper/lower case
		} else {
			if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
			CGRuntime::setLocation(iLocation);
			if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
			return BREAK_INTERRUPTION;
		}
	}
	if (_pVariableToAssign != 0L) {
		DtaScriptVariable* pVariable = visibility.getVariable(*_pVariableToAssign);
		char tcText[] = {'\0', '\0'};
		tcText[0] = (char) _iChar;
		if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
			pVariable->pushItem(tcText);
		} else {
			if (_bConcatVariable) pVariable->concatValue(tcText);
			else pVariable->setValue(tcText);
		}
	}
	if (iImplicitCopyPosition >= 0) {
		register char cChar = (char) _iChar;
		_pBNFScript->writeBinaryData(&cChar, 1);
	}
	return NO_INTERRUPTION;
}

void BNFCharLitteral::compileCpp(CppCompilerEnvironment& theCompilerEnvironment) const {
	int iCursor = theCompilerEnvironment.newCursor();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "// " << toString();
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseLocation_" << iCursor << " = CGRuntime::getLocation();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseImplicitCopy_" << iCursor << " = theEnvironment.skipEmptyChars();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "_compilerClauseSuccess = CGRuntime::readIfEqualTo(";
	theCompilerEnvironment.getBody().writeQuotedChar((char) _iChar);
	theCompilerEnvironment.getBody() << ")";
	if (_bNoCase) {
		theCompilerEnvironment.getBody() << " || CGRuntime::readIfEqualTo(";
		theCompilerEnvironment.getBody().writeQuotedChar((char) (_iChar - 32));
		theCompilerEnvironment.getBody() << ");";
	} else {
		theCompilerEnvironment.getBody() << ";";
	}
	theCompilerEnvironment.getBody().endl();
	if (_bContinue) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) CGRuntime::throwBNFExecutionError(";
		theCompilerEnvironment.getBody().writeString(toString());
		theCompilerEnvironment.getBody() << ");";
		theCompilerEnvironment.getBody().endl();
	} else if (_pVariableToAssign != 0L) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.incrementIndentation();
	} else {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) CGRuntime::setLocation(_compilerClauseLocation_" << iCursor << ");";
		theCompilerEnvironment.getBody().endl();
	}
	if (_pVariableToAssign != 0L) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "char tcText[] = {'\\0', '\\0'};";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "tcText[0] = (char) _iChar;";
		theCompilerEnvironment.getBody().endl();
		if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation();
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".pushItem(tcText);";
			theCompilerEnvironment.getBody().endl();
		} else {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation();
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".setValue(tcText);";
			theCompilerEnvironment.getBody().endl();
		}
		if (!_bContinue) {
			theCompilerEnvironment.decrementIndentation();
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
			theCompilerEnvironment.getBody().endl();
		}
	}
}

std::string BNFCharLitteral::toString() const {
	std::string sString;
	if (_bContinue) sString = "#continue ";
	sString += "'";
	char c = (char) _iChar;
	switch(c) {
		case '\\': sString += "\\\\";break;
		case '\"': sString += "\\\"";break;
		case '\'': sString += "\\'";break;
		case '\a': sString += "\\a";break;
		case '\b': sString += "\\b";break;
		case '\f': sString += "\\f";break;
		case '\n': sString += "\\n";break;
		case '\r': sString += "\\r";break;
		case '\t': sString += "\\t";break;
		case '\v': sString += "\\v";break;
		default:
			sString += c;
	}
	sString += "'";
	return sString + DtaBNFScript::assignmentToString(_pVariableToAssign);
}
