/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif

#include "UtlException.h"
#include "UtlTrace.h"

#ifndef WIN32
#	include "UtlString.h" // for Debian/gcc 2.95.4
#endif

#include "ScpStream.h"
#include "CppCompilerEnvironment.h"
#include "CGRuntime.h"

#include "DtaScriptVariable.h"
#include "ExprScriptVariable.h"
#include "GrfExecutionContext.h"
#include "DtaBNFScript.h"
#include "GrfFunction.h"
#include "BNFClause.h"
#include "BNFClauseCall.h"

namespace CodeWorker {
	class GrfClauseDepth {
	public:
		GrfClauseDepth() {
			GrfFunction::_iCurrentStackDepth++;
			if (GrfFunction::_iCurrentStackDepth > GrfFunction::_iMaxStackDepth) {
				throw UtlException("stack overflow! You may increase the stack reserved for function calls with option '-stack'.");
			}
		}
		inline ~GrfClauseDepth() {
			GrfFunction::_iCurrentStackDepth--;
		}
	};


	BNFClauseCall::BNFClauseCall(DtaBNFScript* pBNFScript, GrfBlock* pParent, bool bContinue, bool bNoCase) : _pBNFScript(pBNFScript), GrfCommand(pParent), _pClause(0L), _pTemplateExpression(0L), _pVariableToAssign(0L), _bContinue(bContinue), _iClauseReturnType(BNFClause::NO_RETURN_TYPE), _bNoCase(bNoCase), _iSuperCallDepth(0) {}

	BNFClauseCall::~BNFClauseCall() {
		for (std::vector<ExprScriptExpression*>::iterator i = _listOfParameters.begin(); i != _listOfParameters.end(); i++) {
			delete *i;
		}
		delete _pTemplateExpression;
		delete _pVariableToAssign;
	}

	bool BNFClauseCall::isABNFCommand() const { return true; }

	void BNFClauseCall::setVariableToAssign(ExprScriptVariable* pVariableToAssign, bool bConcat, BNFClause& theClause) {
		if (pVariableToAssign != 0L) {
			_pVariableToAssign = pVariableToAssign;
			std::string sVariableName = _pVariableToAssign->toString();
			if (sVariableName == theClause.getName()) _iClauseReturnType = theClause.getReturnType();
			_bConcatVariable = bConcat;
		}
	}

	void BNFClauseCall::callBeforeExecutionCBK(DtaScriptVariable& visibility) {
		callRecursiveBeforeExecutionCBK(getCurrentExecutionContext(), visibility);
	}

	void BNFClauseCall::callRecursiveBeforeExecutionCBK(GrfExecutionContext* pContext, DtaScriptVariable& visibility) {
		if (pContext != 0L) {
			callRecursiveBeforeExecutionCBK(pContext->getLastExecutionContext(), visibility);
			pContext->handleStartingBNFClause(this);
			pContext->handleBeforeExecutionCBK(this, visibility);
			pContext = pContext->getLastExecutionContext();
		}
	}

	void BNFClauseCall::callAfterExecutionCBK(DtaScriptVariable& visibility) {
		GrfExecutionContext* pContext = getCurrentExecutionContext();
		while (pContext != 0L) {
			pContext->handleAfterExecutionCBK(this, visibility);
			pContext->handleEndingBNFClause(this);
			pContext = pContext->getLastExecutionContext();
		}
	}

	BNFClause& BNFClauseCall::getInstantiatedClause(DtaScriptVariable& visibility, std::string& sInstantiationKey) const {
		BNFClause* pClause = 0L;
		if (_pTemplateExpression != 0L) {
			sInstantiationKey = _pTemplateExpression->getValue(visibility);
			pClause = _pClause->getInstantiatedClause(sInstantiationKey);
			if (pClause == 0L) throw UtlException("template clause '" +  _pClause->getName() + "<\"" + sInstantiationKey + "\">' hasn't been implemented");
		} else {
			pClause = _pClause;
		}
		return *pClause;
	}

	SEQUENCE_INTERRUPTION_LIST BNFClauseCall::executeInternal(DtaScriptVariable& visibility) {
		std::string sInstantiationKey;
		BNFClause& myClause = getInstantiatedClause(visibility, sInstantiationKey);
		std::string sClauseName = myClause.getName();
		if (!sInstantiationKey.empty()) {
			if (myClause.isGenericKey()) sClauseName += "<" + myClause.getTemplateKey() + "=\"" + sInstantiationKey + "\">";
			else sClauseName += "<\"" + sInstantiationKey + "\">";
		}
		if (_pBNFScript->trace()) {
			char tcMessage[80];
			sprintf(tcMessage, "' [%d]", CGRuntime::getLocation());
			CGRuntime::traceLine(_pBNFScript->traceIndentation() + "IN '" + sClauseName + tcMessage);
			_pBNFScript->incrementTraceIndentation();
			if (!CGRuntime::inputKey(false).empty() || (CGRuntime::getLocation() == 2442) || (CGRuntime::getLocation() == 5250)) {
				char tcMessage[512];
				std::string sFilename = CGRuntime::getInputFilename();
				sprintf(tcMessage, "break in \"%s\", line %d", sFilename.c_str(), CGRuntime::countInputLines());
				CGRuntime::traceLine(tcMessage);
				CGRuntime::traceText("> ");
				std::string sCommand = CGRuntime::inputLine(true);
			}

		}
		UTLTRACE_STACK_FUNCTION(_sParsingFilePtr, sClauseName.c_str(), _iFileLocation);
		GrfClauseDepth newClauseCall;
		if (myClause.getNbCommands() == 0) {
			throw UtlException("clause '" + myClause.getSignature() + "' has never been implemented");
		}
		std::string sName = "##stack## clause '";
		sName += sClauseName + "'";
		DtaScriptVariable stackClause(&visibility, sName.c_str());
		DtaScriptVariable* pReturnValue = stackClause.getNonRecursiveVariable(myClause.getName().c_str());
		if (!sInstantiationKey.empty() && myClause.isGenericKey()) {
			stackClause.setValueAtVariable(myClause.getTemplateKey().c_str(), sInstantiationKey.c_str());
		}
		int iLocation = CGRuntime::getLocation();
		int iNotEmptyLocation = 0; // used only if variable to assign or constant to check
		int iImplicitCopyPosition = -1;
		if (myClause.getPreprocessingIgnoreMode() == UNDEFINED_IGNORE) iImplicitCopyPosition = _pBNFScript->skipEmptyChars(visibility);
		if ((_pVariableToAssign != 0L) || !_listOfConstants.empty()) {
			iNotEmptyLocation = CGRuntime::getLocation();
		}
		if (myClause.isPropagatedParameter()) {
			if (myClause.getParameters().size() < _listOfParameters.size()) {
				char tcExpectedNumber[96];
				sprintf(tcExpectedNumber, "' requires less than %d parameters (%d parameters, propagated)", _listOfParameters.size(), myClause.getParameters().size());
				throw UtlException("call to non-terminal '" + myClause.getSignature() + tcExpectedNumber);
			}
		} else {
			if (myClause.getParameters().size() != _listOfParameters.size()) {
				char tcExpectedNumber[64];
				sprintf(tcExpectedNumber, "' doesn't require %d parameters", _listOfParameters.size());
				throw UtlException("call to non-terminal '" + myClause.getSignature() + tcExpectedNumber);
			}
		}
		std::vector<std::string>::const_iterator j = myClause.getParameters().begin();
		std::vector<EXPRESSION_TYPE>::const_iterator k = myClause.getParameterTypes().begin();
		for (std::vector<ExprScriptExpression*>::const_iterator i = _listOfParameters.begin(); i != _listOfParameters.end(); i++) {
			switch(*k) {
				case VALUE_EXPRTYPE:
					{
						std::string sValue = (*i)->getValue(visibility);
						stackClause.setValueAtVariable(j->c_str(), sValue.c_str());
					}
					break;
				case REFERENCE_EXPRTYPE:
				case VARIABLE_EXPRTYPE:
					{
						register const ExprScriptVariable* pVarExpr = dynamic_cast<const ExprScriptVariable*>(*i);
						if (pVarExpr == 0L) {
							std::string sValue = (*i)->getValue(visibility);
							sValue = CGRuntime::composeCLikeString(sValue);
							throw UtlException("in clause '" + myClause.getSignature() + "', parameter '" + *j + ":node' expects a tree node, not the expression \"" + sValue + "\"");
						}
						DtaScriptVariable* pVariable = visibility.getVariable(*pVarExpr);
						DtaScriptVariable* pParameter = stackClause.createIndex(j->c_str());
						pParameter->setValue(pVariable);
					}
					break;
			}
			j++;
			k++;
		}
		if (myClause.isPropagatedParameter()) {
			while (j != myClause.getParameters().end()) {
				switch(*k) {
					case VALUE_EXPRTYPE:
						{
							DtaScriptVariable* pVariable = visibility.getVariable(j->c_str());
							std::string sValue = pVariable->getValue();
							stackClause.setValueAtVariable(j->c_str(), sValue.c_str());
						}
						break;
					case REFERENCE_EXPRTYPE:
					case VARIABLE_EXPRTYPE:
						{
							DtaScriptVariable* pVariable = visibility.getVariable(j->c_str());
							DtaScriptVariable* pParameter = stackClause.createIndex(j->c_str());
							pParameter->setValue(pVariable);
						}
						break;
				}
				++j;
				++k;
			}
		}
		SEQUENCE_INTERRUPTION_LIST result = myClause.executeClause(stackClause, _iSuperCallDepth);
		if (result == NO_INTERRUPTION) {
			j = myClause.getParameters().begin();
			k = myClause.getParameterTypes().begin();
			for (std::vector<ExprScriptExpression*>::const_iterator i = _listOfParameters.begin(); i != _listOfParameters.end(); i++) {
				if (*k == REFERENCE_EXPRTYPE) {
					DtaScriptVariable* pVariable = visibility.getVariableForReferenceAssignment((const ExprScriptVariable&) (*(*i)));
					DtaScriptVariable* pReference = stackClause.getIndex(j->c_str())->getReferencedVariable();
					pVariable->setValue(pReference);
				}
				j++;
				k++;
			}
			if (myClause.isPropagatedParameter()) {
				while (j != myClause.getParameters().end()) {
					if (*k == REFERENCE_EXPRTYPE) {
						DtaScriptVariable* pVariable = visibility.getVariable(j->c_str(), true, true, false);
						DtaScriptVariable* pReference = stackClause.getIndex(j->c_str())->getReferencedVariable();
						pVariable->setValue(pReference);
					}
					++j;
					++k;
				}
			}
			if (!_listOfConstants.empty()) {
				bool bMatch = false;
				std::string sTextToMatch;
				if (myClause.getReturnType() == BNFClause::VALUE_RETURN_TYPE) sTextToMatch = pReturnValue->getValue();
				else sTextToMatch = CGRuntime::getLastReadChars(CGRuntime::getLocation() - iNotEmptyLocation);
				for (std::vector<std::string>::const_iterator i = _listOfConstants.begin(); i != _listOfConstants.end(); i++) {
					if (_bNoCase) bMatch = (stricmp(i->c_str(), sTextToMatch.c_str()) == 0);
					else bMatch = ((*i) == sTextToMatch);
					if (bMatch) break;
				}
				if (!bMatch) {
					if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
					CGRuntime::setLocation(iLocation);
					if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
					result = BREAK_INTERRUPTION;
				}
			}
			if ((result != BREAK_INTERRUPTION) && (_pVariableToAssign != 0L)) {
				DtaScriptVariable* pVariable = visibility.getVariable(*_pVariableToAssign);
				if (myClause.getReturnType() == BNFClause::LIST_RETURN_TYPE) {
					pVariable->addArrayElements(*pReturnValue);
				} else if (myClause.getReturnType() == BNFClause::NODE_RETURN_TYPE) {
					pVariable->copyAll(*pReturnValue);
				} else if (myClause.getReturnType() == BNFClause::VALUE_RETURN_TYPE) {
					std::string sValue = pReturnValue->getValue();
					if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
						pVariable->pushItem(sValue);
					} else {
						if (_bConcatVariable) pVariable->concatValue(sValue.c_str());
						else pVariable->setValue(sValue.c_str());
					}
				} else {
					std::string sText = CGRuntime::getLastReadChars(CGRuntime::getLocation() - iNotEmptyLocation);
					if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
						pVariable->pushItem(sText);
					} else {
						if (_bConcatVariable) pVariable->concatValue(sText.c_str());
						else pVariable->setValue(sText.c_str());
					}
				}
			}
		} else {
			if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
			CGRuntime::setLocation(iLocation);
			if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
		}
		if (_pBNFScript->trace()) {
			_pBNFScript->decrementTraceIndentation();
			std::string sMessage = _pBNFScript->traceIndentation() + "OUT ";
			if (result == BREAK_INTERRUPTION) sMessage += "fail '";
			else sMessage += "OK '";
			sMessage += sClauseName;
			if (result == BREAK_INTERRUPTION) {
				char tcMessage[80];
				sprintf(tcMessage, "' [%d]", CGRuntime::getLocation());
				sMessage += tcMessage;
			}
			CGRuntime::traceLine(sMessage);
		}
		return result;
	}

	void BNFClauseCall::compileCpp(CppCompilerEnvironment& theCompilerEnvironment) const {
		int iCursor = theCompilerEnvironment.newCursor();
		BNFClause& myClause = getClause();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "// " << toString();
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseLocation = CGRuntime::getLocation();";
		theCompilerEnvironment.getBody().endl();
		if (myClause.getPreprocessingIgnoreMode() == UNDEFINED_IGNORE) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "theEnvironment.skipEmptyChars();";
			theCompilerEnvironment.getBody().endl();
		}
		if ((_pVariableToAssign != 0L) || !_listOfConstants.empty()) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseNotEmptyLocation = CGRuntime::getLocation();";
			theCompilerEnvironment.getBody().endl();
		}
		std::string sClauseName = myClause.getName();
		bool bTemplateDispatching = false;
		if (_pTemplateExpression != 0L) {
			sClauseName = "_compilerTemplateClause_" + sClauseName + "_compilerInstantiation_";
			ExprScriptConstant* pConstant = dynamic_cast<ExprScriptConstant*>(_pTemplateExpression);
			if (pConstant != 0L) {
				sClauseName += pConstant->getConstant();
			} else {
				bTemplateDispatching = true;
			}
		} else if (sClauseName[0] == '#') sClauseName = "_compilerDirectiveClause_" + sClauseName.substr(1);
		else sClauseName = "_compilerClause_" + sClauseName;
		bool bReturnedValue = (!_listOfConstants.empty() && (myClause.getReturnType() == BNFClause::VALUE_RETURN_TYPE)) || ((_pVariableToAssign != 0L) && (myClause.getReturnType() != BNFClause::NO_RETURN_TYPE));
		if (bReturnedValue) {
			theCompilerEnvironment.getBody() << "CppParsingTree_value _compilerClauseReturnedValue;";
			theCompilerEnvironment.getBody().endl();
		}
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "_compilerClauseSuccess = " << DtaScript::convertFilenameAsIdentifier(CppCompilerEnvironment::getRadical(theCompilerEnvironment.getClauseModule(myClause.getName()))) << "::" << sClauseName << "(";
		if (bTemplateDispatching) {
			_pTemplateExpression->compileCppString(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ", ";
		}
		if (bReturnedValue) {
			theCompilerEnvironment.getBody() << "_compilerClauseReturnedValue";
			if (!_listOfParameters.empty()) theCompilerEnvironment.getBody() << ", ";
		}
		int iIndex = 0;
		for (std::vector<ExprScriptExpression*>::const_iterator i = _listOfParameters.begin(); i != _listOfParameters.end(); i++) {
			if (myClause.getParameterType(iIndex) == VALUE_EXPRTYPE) (*i)->compileCpp(theCompilerEnvironment);
			else dynamic_cast<ExprScriptVariable*>(*i)->compileCppForSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ", ";
			iIndex++;
		}
		theCompilerEnvironment.getBody() << "theEnvironment);";
		theCompilerEnvironment.getBody().endl();
		if ((_pVariableToAssign != 0L) || !_listOfConstants.empty()) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
			theCompilerEnvironment.getBody().endl();
			theCompilerEnvironment.incrementIndentation();
			if (!_listOfConstants.empty()) {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "std::string _compilerClauseText_" << iCursor << " = ";
				if (myClause.getReturnType() == BNFClause::VALUE_RETURN_TYPE) {
					theCompilerEnvironment.getBody() << "_compilerClauseReturnedValue.getValue();";
				} else {
					theCompilerEnvironment.getBody() << "CGRuntime::getLastReadChars(CGRuntime::getLocation() - _compilerClauseNotEmptyLocation);";
				}
				theCompilerEnvironment.getBody() << ";";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "_compilerClauseSuccess =";
				for (std::vector<std::string>::const_iterator i = _listOfConstants.begin(); i != _listOfConstants.end(); i++) {
					if (i != _listOfConstants.begin()) theCompilerEnvironment.getBody() << " || ";
					theCompilerEnvironment.getBody() << "(";
					if (_bNoCase) {
						theCompilerEnvironment.getBody() << "stricmp(_compilerClauseText_" << iCursor << ".c_str(), ";
						theCompilerEnvironment.getBody().writeString(*i);
						theCompilerEnvironment.getBody() << ") == 0";
					} else {
						theCompilerEnvironment.getBody() << "_compilerClauseText_" << iCursor << " == ";
						theCompilerEnvironment.getBody().writeString(*i);
					}
					theCompilerEnvironment.getBody() << ")";
				}
				theCompilerEnvironment.getBody() << ";";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) ";
				if (_bContinue) {
					theCompilerEnvironment.getBody() << "CGRuntime::throwBNFExecutionError(";
					theCompilerEnvironment.getBody().writeString(toString());
					theCompilerEnvironment.getBody() << ");";
				} else {
					theCompilerEnvironment.getBody() << "CGRuntime::setLocation(_compilerClauseLocation);";
				}
				theCompilerEnvironment.getBody().endl();
			}
			if (_pVariableToAssign != 0L) {
				if (!_listOfConstants.empty() && !_bContinue) {
					theCompilerEnvironment.incrementIndentation();
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
					theCompilerEnvironment.getBody().endl();
				}
				if (myClause.getReturnType() == BNFClause::LIST_RETURN_TYPE) {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation();
					_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
					theCompilerEnvironment.getBody() << ".addElements(_compilerClauseReturnedValue);";
					theCompilerEnvironment.getBody().endl();
				} else if (myClause.getReturnType() == BNFClause::NODE_RETURN_TYPE) {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation();
					_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
					theCompilerEnvironment.getBody() << ".setAll(_compilerClauseReturnedValue);";
					theCompilerEnvironment.getBody().endl();
				} else {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation();
					if (_listOfConstants.empty()) theCompilerEnvironment.getBody() << "std::string ";
					theCompilerEnvironment.getBody() << "_compilerClauseText_" << iCursor << " = ";
					if (myClause.getReturnType() == BNFClause::VALUE_RETURN_TYPE) {
						theCompilerEnvironment.getBody() << "_compilerClauseReturnedValue.getValue();";
					} else {
						theCompilerEnvironment.getBody() << "CGRuntime::getLastReadChars(CGRuntime::getLocation() - _compilerClauseNotEmptyLocation);";
					}
					theCompilerEnvironment.getBody().endl();
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation();
					_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
					if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
						theCompilerEnvironment.getBody() << ".pushItem(_compilerClauseText_" << iCursor << ");";
						theCompilerEnvironment.getBody().endl();
					} else {
						theCompilerEnvironment.getBody() << ".setValue(_compilerClauseText_" << iCursor << ");";
						theCompilerEnvironment.getBody().endl();
					}
				}
				if (!_listOfConstants.empty() && !_bContinue) {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
					theCompilerEnvironment.getBody().endl();
					theCompilerEnvironment.decrementIndentation();
				}
			}
			theCompilerEnvironment.decrementIndentation();
			if (_bContinue) {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "} else CGRuntime::throwBNFExecutionError(";
				theCompilerEnvironment.getBody().writeString(toString());
				theCompilerEnvironment.getBody() << ");";
			} else {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "} else CGRuntime::setLocation(_compilerClauseLocation);";
			}
			theCompilerEnvironment.getBody().endl();
		} else {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) ";
			if (_bContinue) {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "CGRuntime::throwBNFExecutionError(";
				theCompilerEnvironment.getBody().writeString(toString());
				theCompilerEnvironment.getBody() << ");";
			} else {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "CGRuntime::setLocation(_compilerClauseLocation);";
			}
			theCompilerEnvironment.getBody().endl();
		}
	}

	std::string BNFClauseCall::toString() const {
		std::string sString;
		if (_bContinue) sString = "#continue ";
		sString += getClause().getName();
		if (!getClause().getTemplateKey().empty()) sString += "<\"" + getClause().getTemplateKey() + "\">";
		if (!_listOfParameters.empty()) {
			if (_pTemplateExpression != 0L) sString += "<" + _pTemplateExpression->toString() + ">";
			sString += "(";
			for (std::vector<ExprScriptExpression*>::const_iterator i = _listOfParameters.begin(); i != _listOfParameters.end(); i++) {
				if (i != _listOfParameters.begin()) sString += ", ";
				sString += (*i)->toString();
			}
			sString += ")";
		}
		return sString + DtaBNFScript::constantsToString(_listOfConstants) + DtaBNFScript::assignmentToString(_pVariableToAssign);
	}



	BNFRootClauseCall::BNFRootClauseCall(DtaBNFScript* pBNFScript, GrfBlock* pParent, bool bNoCase) : BNFClauseCall(pBNFScript, pParent, false, bNoCase) {}
	BNFRootClauseCall::~BNFRootClauseCall() {}

	void BNFRootClauseCall::compileCpp(CppCompilerEnvironment& theCompilerEnvironment) const {
		std::string sIgnoreClauseName;
		if (getBNFScript().getIgnoreClause() != 0L) {
			sIgnoreClauseName = getBNFScript().getIgnoreClause()->getName();
			if (sIgnoreClauseName[0] == '#') sIgnoreClauseName = "_compilerDirectiveClause_" + sIgnoreClauseName.substr(1);
			sIgnoreClauseName = DtaScript::convertFilenameAsIdentifier(CppCompilerEnvironment::getRadical(theCompilerEnvironment.getClauseModule(getBNFScript().getIgnoreClause()->getName()))) + "::" + sIgnoreClauseName;
		} else if (getClause().getPreprocessingIgnoreMode() != UNDEFINED_IGNORE) {
			IGNORE_MODE eMode = (IGNORE_MODE) getClause().getPreprocessingIgnoreMode();
			if (eMode == NOT_IGNORE) {
				sIgnoreClauseName = "0L";
			} else {
				if (eMode == IGNORE_CLAUSE) sIgnoreClauseName = "_compilerDirectiveClause_" + sIgnoreClauseName.substr(1);
				sIgnoreClauseName = DtaScript::convertFilenameAsIdentifier(CppCompilerEnvironment::getRadical(theCompilerEnvironment.getClauseModule(getBNFScript().getIgnoreClause()->getName()))) + "::" + sIgnoreClauseName;
			}
		} else {
			sIgnoreClauseName = "0L";
		}
		std::string sClauseName = getClause().getName();
		if (getClause().isATemplateInstantiation()) sClauseName = "_compilerTemplateClause_" + sClauseName + "_compilerInstantiation_" + theCompilerEnvironment.convertTemplateKey(getClause().getTemplateKey());
		else if (sClauseName[0] == '#') sClauseName = "_compilerDirectiveClause_" + sClauseName.substr(1);
		else sClauseName = "_compilerClause_" + sClauseName;
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "CGBNFRuntimeEnvironment theEnvironment(" << sIgnoreClauseName << ", \"" << getBNFScript().getIgnoreMode() << "\", false);";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << DtaScript::convertFilenameAsIdentifier(CppCompilerEnvironment::getRadical(theCompilerEnvironment.getClauseModule(getClause().getName()))) << "::" << sClauseName << "(";
		int iIndex = 0;
		for (std::vector<ExprScriptExpression*>::const_iterator i = getParameters().begin(); i != getParameters().end(); i++) {
			if (getClause().getParameterType(iIndex) == VALUE_EXPRTYPE) (*i)->compileCpp(theCompilerEnvironment);
			else dynamic_cast<ExprScriptVariable*>(*i)->compileCppForSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ", ";
			iIndex++;
		}
		theCompilerEnvironment.getBody() << "theEnvironment);";
		theCompilerEnvironment.getBody().endl();
	}
}
