/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif

#include "UtlException.h"
#include "ScpStream.h"
#include "CppCompilerEnvironment.h"
#include "CGRuntime.h"

#include "DtaScriptVariable.h"
#include "ExprScriptVariable.h"
#include "DtaBNFScript.h"
#include "BNFClause.h"
#include "BNFMultiplicity.h" // for 'BNFMultiplicityBoundaries'
#include "BNFFindToken.h"

namespace CodeWorker {

	BNFFindToken::BNFFindToken(DtaBNFScript* pBNFScript, GrfBlock* pParent, bool bContinue) : BNFStepper(pBNFScript, pParent, bContinue), _pIntermediateSequence(0L), _pIntermediateVariableToAssign(0L), _pFinalVariableToAssign(0L), _pBoundaries(0L) {}

	BNFFindToken::~BNFFindToken() {
		delete _pIntermediateSequence;
		delete _pIntermediateVariableToAssign;
		delete _pFinalVariableToAssign;
		delete _pBoundaries;
	}

	GrfBlock& BNFFindToken::createIntermediateSequence() {
		_pIntermediateSequence = new GrfBlock(this);
		return *_pIntermediateSequence;
	}

	void BNFFindToken::setIntermediateVariableToAssign(ExprScriptVariable* pVariableToAssign, bool bConcat, BNFClause& theClause) {
		if (pVariableToAssign != 0L) {
			_pIntermediateVariableToAssign = pVariableToAssign;
			std::string sVariableName = _pIntermediateVariableToAssign->toString();
			if (sVariableName == theClause.getName()) _iClauseReturnType = theClause.getReturnType();
			_bConcatIntermediateVariable = bConcat;
		}
	}

	void BNFFindToken::setFinalVariableToAssign(ExprScriptVariable* pVariableToAssign, bool bConcat, BNFClause& theClause) {
		if (pVariableToAssign != 0L) {
			_pFinalVariableToAssign = pVariableToAssign;
			std::string sVariableName = _pFinalVariableToAssign->toString();
			if (sVariableName == theClause.getName()) _iClauseReturnType = theClause.getReturnType();
			_bConcatFinalVariable = bConcat;
		}
	}

	SEQUENCE_INTERRUPTION_LIST BNFFindToken::executeInternal(DtaScriptVariable& visibility) {
		SEQUENCE_INTERRUPTION_LIST result;
		int iLocation = CGRuntime::getLocation();
		int iImplicitCopyPosition = _pBNFScript->skipEmptyChars(visibility);
		int iNotEmptyLocation = CGRuntime::getLocation();
		int iIntermediateLocation = iNotEmptyLocation; // for 'intermediate code' only
		_iStepLocation = 0;
		int iBegin;
		int iEnd;
		if (_pBoundaries != 0L) {
			_pBoundaries->computeBoundaries(visibility, iBegin, iEnd);
		} else {
			iBegin = 0;
			iEnd = -1;
		}
		int iLoop;
		for (iLoop = 0; (iLoop <= iBegin) || (iEnd < 0) || (iLoop <= iEnd); iLoop++) {
			_pBNFScript->skipEmptyChars(visibility);
			result = GrfBlock::executeInternal(visibility);
			if (result != BREAK_INTERRUPTION) break;
			if (_iStepLocation > iNotEmptyLocation) {
				CGRuntime::setLocation(_iStepLocation);
				iIntermediateLocation = _iStepLocation;
				if (iImplicitCopyPosition >= 0) {
					std::string sText = CGRuntime::getLastReadChars(_iStepLocation - iNotEmptyLocation);
					_pBNFScript->writeBinaryData(sText.c_str(), sText.size());
				}
			} else {
				int iChar = CGRuntime::readCharAsInt();
				if (iChar < 0) break;
				iIntermediateLocation++;
				if (iImplicitCopyPosition >= 0) {
					char cChar = (char) iChar;
					_pBNFScript->writeBinaryData(&cChar, 1);
				}
			}
		}
		if (result == NO_INTERRUPTION) {
			if (iLoop >= iBegin) {
				int iCurrentLocation = CGRuntime::getLocation();
				result = resolveAssignment(visibility, _listOfConstants, _pVariableToAssign, _bConcatVariable, iImplicitCopyPosition, iLocation, iNotEmptyLocation, iCurrentLocation);
				if (result == NO_INTERRUPTION) {
					result = resolveAssignment(visibility, _listOfFinalConstants, _pFinalVariableToAssign, _bConcatFinalVariable, iImplicitCopyPosition, iLocation, iIntermediateLocation, iCurrentLocation);
					if (result == NO_INTERRUPTION) {
						CGRuntime::setLocation(iIntermediateLocation);
						result = resolveAssignment(visibility, _listOfIntermediateConstants, _pIntermediateVariableToAssign, _bConcatIntermediateVariable, iImplicitCopyPosition, iLocation, iNotEmptyLocation, iIntermediateLocation);
						if ((_pIntermediateSequence != 0L) && (result == NO_INTERRUPTION)) {
							result = _pIntermediateSequence->execute(visibility);
						}
						CGRuntime::setLocation(iCurrentLocation);
					}
				}
			} else {
				result = BREAK_INTERRUPTION;
			}
			if ((result != NO_INTERRUPTION) && (result != CONTINUE_INTERRUPTION)) {
				CGRuntime::setLocation(iLocation);
				if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
				if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
			}
		} else if (result != CONTINUE_INTERRUPTION) {
			CGRuntime::setLocation(iLocation);
			if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
			if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
		}
		return result;
	}

	SEQUENCE_INTERRUPTION_LIST BNFFindToken::resolveAssignment(DtaScriptVariable& visibility, const std::vector<std::string>& listOfConstants, ExprScriptVariable* pVariableToAssign, bool bConcatVariable, int iImplicitCopyPosition, int iOrgLocation, int iBeginLocation, int iEndLocation) {
		std::string sTextToMatch;
		if (!listOfConstants.empty()) {
			bool bMatch = false;
			sTextToMatch = CGRuntime::getLastReadChars(iEndLocation - iBeginLocation);
			for (std::vector<std::string>::const_iterator i = listOfConstants.begin(); i != listOfConstants.end(); i++) {
				bMatch = ((*i) == sTextToMatch);
				if (bMatch) break;
			}
			if (!bMatch) {
				if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
				CGRuntime::setLocation(iOrgLocation);
				if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
				return BREAK_INTERRUPTION;
			}
		}
		if (pVariableToAssign != 0L) {
			DtaScriptVariable* pVariable = visibility.getVariable(*pVariableToAssign);
			if (sTextToMatch.empty()) sTextToMatch = CGRuntime::getLastReadChars(iEndLocation - iBeginLocation);
			if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
				pVariable->pushItem(sTextToMatch);
			} else {
				if (bConcatVariable) pVariable->concatValue(sTextToMatch.c_str());
				else pVariable->setValue(sTextToMatch.c_str());
			}
		}
		return NO_INTERRUPTION;
	}

	void BNFFindToken::compileCpp(CppCompilerEnvironment& theCompilerEnvironment) const {
		std::vector<GrfCommand*>::const_iterator i = getCommands().begin();
		if (i != getCommands().end()) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "// " << toString();
			theCompilerEnvironment.getBody().endl();
			int iCursor = theCompilerEnvironment.newCursor();
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseLocation_" << iCursor << " = CGRuntime::getLocation();";
			theCompilerEnvironment.getBody().endl();
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseImplicitCopy_" << iCursor << " = theEnvironment.skipEmptyChars();";
			theCompilerEnvironment.getBody().endl();
			if (_pVariableToAssign != 0L) {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseNotEmptyLocation_" << iCursor << " = CGRuntime::getLocation();";
				theCompilerEnvironment.getBody().endl();
			}
			int iOldBNFStepperCursor = theCompilerEnvironment.getBNFStepperCursor();
			theCompilerEnvironment.setBNFStepperCursor(iCursor);
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseNextLocation_" << iCursor << " = 0;";
			theCompilerEnvironment.getBody().endl();
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "do {";
			theCompilerEnvironment.getBody().endl();
			theCompilerEnvironment.incrementIndentation();
			for (;;) {
				(*i)->compileCpp(theCompilerEnvironment);
				i++;
				if (i == getCommands().end()) break;
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.incrementIndentation();
			}
			int j = getNbCommands() - 1;
			while (j > 0) {
				theCompilerEnvironment.decrementIndentation();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
				theCompilerEnvironment.getBody().endl();
				j--;
			}
			theCompilerEnvironment.decrementIndentation();
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "} while (!_compilerClauseSuccess && !CGRuntime::readChar().empty());";
			theCompilerEnvironment.getBody().endl();
			if (_bContinue) {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) CGRuntime::throwBNFExecutionError(";
				theCompilerEnvironment.getBody().writeString(toString());
				theCompilerEnvironment.getBody() << ");";
				theCompilerEnvironment.getBody().endl();
			}
			if (_pVariableToAssign != 0L) {
				if (!_bContinue) {
					theCompilerEnvironment.incrementIndentation();
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
					theCompilerEnvironment.getBody().endl();
				}
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tstd::string sText = CGRuntime::getLastReadChars(CGRuntime::getLocation() - _compilerClauseNotEmptyLocation_" << iCursor << ");";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
				_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
				if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
					theCompilerEnvironment.getBody() << ".pushItem(sText);";
				} else {
					theCompilerEnvironment.getBody() << ".setValue(sText);";
				}
				theCompilerEnvironment.getBody().endl();
				if (!_bContinue) {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
					theCompilerEnvironment.getBody().endl();
					theCompilerEnvironment.decrementIndentation();
				}
			}
			theCompilerEnvironment.setBNFStepperCursor(iOldBNFStepperCursor);
		}
	}

	std::string BNFFindToken::toString() const {
		std::string sText = "->";
		if (_pBoundaries != 0L) {
			sText += _pBoundaries->toString();
			if (sText.size() == 2) sText += "1";
		}
		if (hasIntermediateCode()) {
			std::string sConstVar = DtaBNFScript::constantsToString(_listOfIntermediateConstants) + DtaBNFScript::assignmentToString(_pIntermediateVariableToAssign);
			if (!_listOfFinalConstants.empty() || (_pFinalVariableToAssign != 0L)) {
				sConstVar += "-" + DtaBNFScript::constantsToString(_listOfFinalConstants) + DtaBNFScript::assignmentToString(_pFinalVariableToAssign);
			}
			sText += "(" + sConstVar;
			if (_pIntermediateSequence != 0L) {
				if (!sConstVar.empty()) sText += " ";
				sText += _pIntermediateSequence->toString();
			}
			sText += ")";
		}
		sText += getCommands()[0]->toString() + DtaBNFScript::constantsToString(_listOfConstants) + DtaBNFScript::assignmentToString(_pVariableToAssign);
		if (_bContinue) sText = "#continue " + sText;
		return sText;
	}
}
