/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif


#include "UtlException.h"
#include "ScpStream.h"
#include "CppCompilerEnvironment.h"
#include "CGRuntime.h"

#include "DtaScriptVariable.h"
#include "ExprScriptVariable.h"
#include "DtaBNFScript.h"
#include "BNFClause.h"
#include "BNFMultiplicity.h"

namespace CodeWorker {

	BNFMultiplicityBoundaries::~BNFMultiplicityBoundaries() {
		if (!bConstantBoundaries) {
			delete choice.variable._pBegin;
			delete choice.variable._pEnd;
		}
	}

	void BNFMultiplicityBoundaries::setMultiplicity(int iBegin, int iEnd) {
		bConstantBoundaries = true;
		choice.constant._iBegin = iBegin;
		choice.constant._iEnd = iEnd;
	}

	void BNFMultiplicityBoundaries::setMultiplicity(ExprScriptExpression* pBegin, ExprScriptExpression* pEnd) {
		bConstantBoundaries = false;
		choice.variable._pBegin = pBegin;
		choice.variable._pEnd = pEnd;
	}

	void BNFMultiplicityBoundaries::computeBoundaries(DtaScriptVariable& visibility, int& iBegin, int& iEnd) const {
		if (bConstantBoundaries) {
			iBegin = choice.constant._iBegin;
			iEnd   = choice.constant._iEnd;
		} else {
			iBegin = choice.variable._pBegin->getIntValue(visibility);
			if (choice.variable._pEnd == 0L) iEnd = iBegin;
			else iEnd = choice.variable._pEnd->getIntValue(visibility);
		}
	}

	std::string BNFMultiplicityBoundaries::toString() const {
		std::string sString;
		if (bConstantBoundaries) {
			if (choice.constant._iBegin == 0) {
				if (choice.constant._iEnd < 0) sString = "*";
				else if (choice.constant._iEnd == 1) sString = "?";
				else {
					char tcNumber[32];
					sprintf(tcNumber, "0..%d", choice.constant._iEnd);
					sString = tcNumber;
				}
			} else if (choice.constant._iBegin == 1) {
				if (choice.constant._iEnd < 0) sString = "+";
				else if (choice.constant._iEnd != 1) {
					char tcNumber[32];
					sprintf(tcNumber, "1..%d", choice.constant._iEnd);
					sString = tcNumber;
				}
			} else {
				if (choice.constant._iBegin == choice.constant._iEnd) {
					char tcNumber[32];
					sprintf(tcNumber, "%d", choice.constant._iBegin);
					sString = tcNumber;
				} else {
					char tcNumber[32];
					sprintf(tcNumber, "%d..%d", choice.constant._iBegin, choice.constant._iEnd);
					sString = tcNumber;
				}
			}
		} else {
			sString = "#repeat(" + choice.variable._pBegin->toString();
			if (choice.variable._pEnd != 0L) {
				sString += ", " + choice.variable._pEnd->toString();
			}
			sString += ")";
		}
		return sString;
	}


	BNFMultiplicity::BNFMultiplicity(DtaBNFScript* pBNFScript, GrfBlock* pParent, bool bContinue) : _pBNFScript(pBNFScript), GrfBlock(pParent), _pVariableToAssign(0L), _bContinue(bContinue), _iClauseReturnType(BNFClause::NO_RETURN_TYPE) {}

	BNFMultiplicity::~BNFMultiplicity() {
		delete _pVariableToAssign;
	}

	bool BNFMultiplicity::isABNFCommand() const { return true; }

	void BNFMultiplicity::setVariableToAssign(ExprScriptVariable* pVariableToAssign, bool bConcat, BNFClause& theClause) {
		if (pVariableToAssign != 0L) {
			_pVariableToAssign = pVariableToAssign;
			std::string sVariableName = _pVariableToAssign->toString();
			if (sVariableName == theClause.getName()) _iClauseReturnType = theClause.getReturnType();
			_bConcatVariable = bConcat;
		}
	}

	SEQUENCE_INTERRUPTION_LIST BNFMultiplicity::execute(DtaScriptVariable& visibility) {
		SEQUENCE_INTERRUPTION_LIST result = NO_INTERRUPTION;
		int iFirstLocation = CGRuntime::getLocation();
		int iImplicitCopyPosition = -1;
		int iNotEmptyLocation;
		if (!_listOfConstants.empty() || (_pVariableToAssign != 0L)) {
			iImplicitCopyPosition = _pBNFScript->skipEmptyChars(visibility);
			iNotEmptyLocation = CGRuntime::getLocation();
		}
		int iLastLocation = iFirstLocation;
		int iBegin;
		int iEnd;
		_boundaries.computeBoundaries(visibility, iBegin, iEnd);
		int i;
		for (i = 0; i < iBegin; i++) {
			result = GrfBlock::executeInternal(visibility);
			if (result != NO_INTERRUPTION) {
				if (result == CONTINUE_INTERRUPTION) {
					if (i == iBegin - 1) result = NO_INTERRUPTION;
					else result = BREAK_INTERRUPTION;
				}
				break;
			}
			iLastLocation = CGRuntime::getLocation();
		}
		if (result == NO_INTERRUPTION) {
			int iInfiniteLoopCounter = 0;
			for (i = iEnd - iBegin; i != 0; i--) {
				result = GrfBlock::executeInternal(visibility);
				if (result != NO_INTERRUPTION) {
					if (result != CONTINUE_INTERRUPTION) {
						CGRuntime::setLocation(iLastLocation);
						result = NO_INTERRUPTION;
					}
					break;
				}
				register int iNewLocation = CGRuntime::getLocation();
	/*			if (iLastLocation == iNewLocation) {
					iInfiniteLoopCounter++;
					if (iInfiniteLoopCounter > 10) throw UtlException("infinite loop expected: the parse cursor hasn't move after repeating a sequence 10 times");
				}
	*/			iLastLocation = iNewLocation;
			}
		}
		if (result == NO_INTERRUPTION) {
			if (!_listOfConstants.empty()) {
				bool bMatch = false;
				std::string sTextToMatch = CGRuntime::getLastReadChars(CGRuntime::getLocation() - iNotEmptyLocation);
				for (std::vector<std::string>::const_iterator i = _listOfConstants.begin(); i != _listOfConstants.end(); i++) {
					bMatch = ((*i) == sTextToMatch);
					if (bMatch) break;
				}
				if (!bMatch) {
					if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
					CGRuntime::setLocation(iFirstLocation);
					if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
					return BREAK_INTERRUPTION;
				}
			}
			if (_pVariableToAssign != 0L) {
				DtaScriptVariable* pVariable = visibility.getVariable(*_pVariableToAssign);
				std::string sText = CGRuntime::getLastReadChars(CGRuntime::getLocation() - iNotEmptyLocation);
				if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
					pVariable->pushItem(sText);
				} else {
					if (_bConcatVariable) pVariable->concatValue(sText.c_str());
					else pVariable->setValue(sText.c_str());
				}
			}
		} else if (result == CONTINUE_INTERRUPTION) {
			if ((iBegin != 1) || (iEnd != 1)) result = NO_INTERRUPTION;
		} else {
			if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
			CGRuntime::setLocation(iFirstLocation);
			if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
		}
		return result;
	}

	void BNFMultiplicity::compileCpp(CppCompilerEnvironment& theCompilerEnvironment) const {
		if (!_boundaries.bConstantBoundaries) throw UtlException("BNFMultiplicity::compileCpp(): variable boundaries not handled yet!");
		std::vector<GrfCommand*>::const_iterator i = getCommands().begin();
		if (i != getCommands().end()) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "// " << toString();
			theCompilerEnvironment.getBody().endl();
			int iCursor = theCompilerEnvironment.newCursor();
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseCursor_" << iCursor << " = CGRuntime::getLocation();";
			theCompilerEnvironment.getBody().endl();
			if (_pVariableToAssign != 0L) {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "theEnvironment.skipEmptyChars();";
				theCompilerEnvironment.getBody().endl();
			}
			if (_boundaries.choice.constant._iEnd != 1) {
				if ((_boundaries.choice.constant._iBegin > 0) || (_boundaries.choice.constant._iEnd > 0)) {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClause_iteration_" << iCursor << " = 0;";
					theCompilerEnvironment.getBody().endl();
				}
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "do {";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.incrementIndentation();
			}
			for (;;) {
				(*i)->compileCpp(theCompilerEnvironment);
				i++;
				if (i == getCommands().end()) break;
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.incrementIndentation();
			}
			int j = getNbCommands() - 1;
			while (j > 0) {
				theCompilerEnvironment.decrementIndentation();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
				theCompilerEnvironment.getBody().endl();
				j--;
			}
			if (_boundaries.choice.constant._iEnd != 1) {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) {";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	CGRuntime::setLocation(_compilerClauseCursor_" << iCursor << ");";
				theCompilerEnvironment.getBody().endl();
				if (_boundaries.choice.constant._iBegin == 0) {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	_compilerClauseSuccess = true;";
					theCompilerEnvironment.getBody().endl();
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	break;";
					theCompilerEnvironment.getBody().endl();
				}
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "} else {";
				theCompilerEnvironment.getBody().endl();
				if ((_boundaries.choice.constant._iBegin > 0) || (_boundaries.choice.constant._iEnd > 0)) {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	_compilerClause_iteration_" << iCursor << "++;";
					theCompilerEnvironment.getBody().endl();
				}
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	";
				if (_boundaries.choice.constant._iBegin > 1) theCompilerEnvironment.getBody() << "if (_compilerClause_iteration_" << iCursor << " >= " << _boundaries.choice.constant._iBegin << ") ";
				theCompilerEnvironment.getBody() << "_compilerClauseCursor_" << iCursor << " = CGRuntime::getLocation();";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.decrementIndentation();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "} while (_compilerClauseSuccess";
				if (_boundaries.choice.constant._iEnd > 0) theCompilerEnvironment.getBody() << " && (_compilerClause_iteration_" << iCursor << " < " << _boundaries.choice.constant._iEnd << ")";
				theCompilerEnvironment.getBody() << ");";
				theCompilerEnvironment.getBody().endl();
				if (_boundaries.choice.constant._iBegin > 0) {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess && (_compilerClause_iteration_" << iCursor << " >= " << _boundaries.choice.constant._iBegin << ")) _compilerClauseSuccess = true;";
					theCompilerEnvironment.getBody().endl();
				}
			} else if (_boundaries.choice.constant._iBegin == 0) {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) {";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	CGRuntime::setLocation(_compilerClauseCursor_" << iCursor << ");";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "	_compilerClauseSuccess = true;";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
				theCompilerEnvironment.getBody().endl();
			} else {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) CGRuntime::setLocation(_compilerClauseCursor_" << iCursor << ");";
				theCompilerEnvironment.getBody().endl();
			}
			if (_bContinue) {
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) CGRuntime::throwBNFExecutionError(";
				theCompilerEnvironment.getBody().writeString(toString());
				theCompilerEnvironment.getBody() << ");";
				theCompilerEnvironment.getBody().endl();
			}
			if (_pVariableToAssign != 0L) {
				if (!_bContinue) {
					theCompilerEnvironment.incrementIndentation();
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
					theCompilerEnvironment.getBody().endl();
				}
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tstd::string sText = CGRuntime::getLastReadChars(CGRuntime::getLocation() - _compilerClauseCursor_" << iCursor << ");";
				theCompilerEnvironment.getBody().endl();
				theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
				_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
				if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
					theCompilerEnvironment.getBody() << ".pushItem(sText);";
				} else {
					theCompilerEnvironment.getBody() << ".setValue(sText);";
				}
				theCompilerEnvironment.getBody().endl();
				if (!_bContinue) {
					theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
					theCompilerEnvironment.getBody().endl();
					theCompilerEnvironment.decrementIndentation();
				}
			}
		}
	}

	std::string BNFMultiplicity::toString() const {
		std::string sString;
		if (_bContinue) sString = "#continue ";
		sString += "[";
		for (std::vector<GrfCommand*>::const_iterator i = getCommands().begin(); i != getCommands().end(); ++i) {
			if (i != getCommands().begin()) sString += " ";
			sString += (*i)->toString();
		}
		sString += "]";
		sString += _boundaries.toString();
		return sString + DtaBNFScript::assignmentToString(_pVariableToAssign);
	}
}
