/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif

#include "UtlException.h"

#ifndef WIN32
#	include "UtlString.h" // for Debian/gcc 2.95.4
#endif

#include "ScpStream.h"
#include "CppCompilerEnvironment.h"
#include "CGRuntime.h"

#include "DtaScriptVariable.h"
#include "ExprScriptVariable.h"
#include "DtaBNFScript.h"
#include "BNFClause.h"
#include "BNFReadCString.h"

using namespace CodeWorker;

BNFReadCString::BNFReadCString(DtaBNFScript* pBNFScript, GrfBlock* pParent, bool bContinue, bool bNoCase) : _pBNFScript(pBNFScript), GrfCommand(pParent), _pVariableToAssign(0L), _bContinue(bContinue), _iClauseReturnType(BNFClause::NO_RETURN_TYPE), _bNoCase(bNoCase) {}

BNFReadCString::~BNFReadCString() {
	delete _pVariableToAssign;
}

bool BNFReadCString::isABNFCommand() const { return true; }

void BNFReadCString::setVariableToAssign(ExprScriptVariable* pVariableToAssign, bool bConcat, BNFClause& theClause) {
	if (pVariableToAssign != 0L) {
		_pVariableToAssign = pVariableToAssign;
		std::string sVariableName = _pVariableToAssign->toString();
		if (sVariableName == theClause.getName()) _iClauseReturnType = theClause.getReturnType();
		_bConcatVariable = bConcat;
	}
}

SEQUENCE_INTERRUPTION_LIST BNFReadCString::executeInternal(DtaScriptVariable& visibility) {
	bool bSuccess;
	int iLocation = CGRuntime::getLocation();
	int iImplicitCopyPosition = _pBNFScript->skipEmptyChars(visibility);
	int iNotEmptyLocation = CGRuntime::getLocation();
	std::string sValue;
	bSuccess = CGRuntime::readString(sValue);
	if (!bSuccess) {
		if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
		CGRuntime::setLocation(iLocation);
		if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
		return BREAK_INTERRUPTION;
	}
	if (!_listOfConstants.empty()) {
		bool bMatch = false;
		for (std::vector<std::string>::const_iterator i = _listOfConstants.begin(); i != _listOfConstants.end(); i++) {
			if (_bNoCase) bMatch = (stricmp(i->c_str(), sValue.c_str()) == 0);
			else bMatch = (*i == sValue);
			if (bMatch) break;
		}
		if (!bMatch) {
			if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
			CGRuntime::setLocation(iLocation);
			if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
			return BREAK_INTERRUPTION;
		}
	}
	if (_pVariableToAssign != 0L) {
		DtaScriptVariable* pVariable = visibility.getVariable(*_pVariableToAssign);
		if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
			pVariable->pushItem(sValue);
		} else {
			if (_bConcatVariable) pVariable->concatValue(sValue.c_str());
			else pVariable->setValue(sValue.c_str());
		}
	}
	if (iImplicitCopyPosition >= 0) {
		std::string sText = CGRuntime::getLastReadChars(CGRuntime::getLocation() - iNotEmptyLocation);
		_pBNFScript->writeBinaryData(sText.c_str(), sText.size());
	}
	return NO_INTERRUPTION;
}

void BNFReadCString::compileCpp(CppCompilerEnvironment& theCompilerEnvironment) const {
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "// " << toString();
	theCompilerEnvironment.getBody().endl();
	int iCursor = theCompilerEnvironment.newCursor();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseLocation_" << iCursor << " = CGRuntime::getLocation();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseImplicitCopy_" << iCursor << " = theEnvironment.skipEmptyChars();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "std::string _compilerClauseText_" << iCursor << ";";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "_compilerClauseSuccess = CGRuntime::readString(_compilerClauseText_" << iCursor << ");";
	theCompilerEnvironment.getBody().endl();
	if (!_listOfConstants.empty()) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t_compilerClauseSuccess =";
		for (std::vector<std::string>::const_iterator i = _listOfConstants.begin(); i != _listOfConstants.end(); i++) {
			if (i != _listOfConstants.begin()) theCompilerEnvironment.getBody() << " || ";
			theCompilerEnvironment.getBody() << "(";
			if (_bNoCase) {
				theCompilerEnvironment.getBody() << "stricmp(_compilerClauseText_" << iCursor << ".c_str(), ";
				theCompilerEnvironment.getBody().writeString(*i);
				theCompilerEnvironment.getBody() << ") == 0";
			} else {
				theCompilerEnvironment.getBody() << "_compilerClauseText_" << iCursor << " == ";
				theCompilerEnvironment.getBody().writeString(*i);
			}
			theCompilerEnvironment.getBody() << ")";
		}
		theCompilerEnvironment.getBody() << ";";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
		theCompilerEnvironment.getBody().endl();
	}
	if (_pVariableToAssign != 0L) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
		theCompilerEnvironment.getBody().endl();
		if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".pushItem(_compilerClauseText_" << iCursor << ");";
			theCompilerEnvironment.getBody().endl();
		} else {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".setValue(_compilerClauseText_" << iCursor << ");";
			theCompilerEnvironment.getBody().endl();
		}
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
		theCompilerEnvironment.getBody().endl();
	}
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) {";
	theCompilerEnvironment.getBody().endl();
	if (_bContinue) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::throwBNFExecutionError(";
		theCompilerEnvironment.getBody().writeString(toString());
		theCompilerEnvironment.getBody() << ");";
	} else {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::setLocation(_compilerClauseLocation_" << iCursor << ");";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tif (_compilerClauseImplicitCopy_" << iCursor << " >= 0) CGRuntime::resizeOutputStream(_compilerClauseImplicitCopy_" << iCursor << ");";
	}
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "} else if (_compilerClauseImplicitCopy_" << iCursor << " >= 0) CGRuntime::writeBinaryData(_compilerClauseText_" << iCursor << ".c_str(), _compilerClauseText_" << iCursor << ".size());";
	theCompilerEnvironment.getBody().endl();
}

std::string BNFReadCString::toString() const {
	std::string sText = "#readCString" + DtaBNFScript::constantsToString(_listOfConstants) + DtaBNFScript::assignmentToString(_pVariableToAssign);
	if (_bContinue) sText = "#continue " + sText;
	return sText;
}
