/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif

#include "UtlException.h"
#include "ScpStream.h"
#include "CppCompilerEnvironment.h"
#include "CGRuntime.h"

#include "DtaScriptVariable.h"
#include "ExprScriptVariable.h"
#include "DtaBNFScript.h"
#include "BNFClause.h"
#include "BNFReadChar.h"

using namespace CodeWorker;

BNFReadChar::BNFReadChar(DtaBNFScript* pBNFScript, GrfBlock* pParent, bool bContinue) : _pBNFScript(pBNFScript), GrfCommand(pParent), _pVariableToAssign(0L), _bContinue(bContinue), _iClauseReturnType(BNFClause::NO_RETURN_TYPE) {}

BNFReadChar::~BNFReadChar() {
	delete _pVariableToAssign;
}

bool BNFReadChar::isABNFCommand() const { return true; }

void BNFReadChar::setVariableToAssign(ExprScriptVariable* pVariableToAssign, bool bConcat, BNFClause& theClause) {
	if (pVariableToAssign != 0L) {
		_pVariableToAssign = pVariableToAssign;
		std::string sVariableName = _pVariableToAssign->toString();
		if (sVariableName == theClause.getName()) _iClauseReturnType = theClause.getReturnType();
		_bConcatVariable = bConcat;
	}
}

void BNFReadChar::setConstantsToMatch(const std::vector<std::string>& listOfConstants) {
	_listOfConstants = listOfConstants;
}

SEQUENCE_INTERRUPTION_LIST BNFReadChar::executeInternal(DtaScriptVariable& visibility) {
	int iLocation = CGRuntime::getLocation();
	int iImplicitCopyPosition = _pBNFScript->skipEmptyChars(visibility);
	int iChar = CGRuntime::readCharAsInt();
	if (iChar < 0) {
		if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
		CGRuntime::setLocation(iLocation);
		if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
		return BREAK_INTERRUPTION;
	}
	if (!_listOfConstants.empty()) {
		bool bMatch = false;
		for (std::vector<std::string>::const_iterator i = _listOfConstants.begin(); i != _listOfConstants.end(); i++) {
			bMatch = ((int) ((*i)[0]) == iChar);
			if (bMatch) break;
		}
		if (!bMatch) {
			if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
			CGRuntime::setLocation(iLocation);
			if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
			return BREAK_INTERRUPTION;
		}
	}
	if (_pVariableToAssign != 0L) {
		DtaScriptVariable* pVariable = visibility.getVariable(*_pVariableToAssign);
		postExecutionAssignment(*pVariable, (unsigned char) iChar);
	}
	if (iImplicitCopyPosition >= 0) {
		register char cChar = (char) iChar;
		_pBNFScript->writeBinaryData((const char*) &cChar, 1);
	}
	return NO_INTERRUPTION;
}

void BNFReadChar::postExecutionAssignment(DtaScriptVariable& variable, unsigned char cChar) {
	char tcText[] = {'\0', '\0'};
	tcText[0] = cChar;
	if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
		variable.pushItem(tcText);
	} else {
		if (_bConcatVariable) variable.concatValue(tcText);
		else variable.setValue(tcText);
	}
}

void BNFReadChar::compileCpp(CppCompilerEnvironment& theCompilerEnvironment) const {
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "// " << toString();
	theCompilerEnvironment.getBody().endl();
	int iCursor = theCompilerEnvironment.newCursor();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseLocation_" << iCursor << " = CGRuntime::getLocation();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseImplicitCopy_" << iCursor << " = theEnvironment.skipEmptyChars();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseChar_" << iCursor << " = CGRuntime::readCharAsInt();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseChar_" << iCursor << " < 0) {";
	theCompilerEnvironment.getBody().endl();
	if (_bContinue) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::throwBNFExecutionError(";
		theCompilerEnvironment.getBody().writeString(toString());
		theCompilerEnvironment.getBody() << ");";
	} else {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t_compilerClauseSuccess = false;";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::setLocation(_compilerClauseLocation" << iCursor << ");";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tif (_compilerClauseImplicitCopy_" << iCursor << " >= 0) CGRuntime::resizeOutputStream(_compilerClauseImplicitCopy_" << iCursor << ");";
	}
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "else {";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t_compilerClauseSuccess = true;";
	theCompilerEnvironment.getBody().endl();
	if (!_listOfConstants.empty()) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tswitch(_compilerClauseChar_" << iCursor << ") {";
		theCompilerEnvironment.getBody().endl();
		for (std::vector<std::string>::const_iterator i = _listOfConstants.begin(); i != _listOfConstants.end(); i++) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t\tcase " << (int) ((unsigned char) ((*i)[0])) << ": // '" << CGRuntime::composeCLikeString(*i) << "'";
			theCompilerEnvironment.getBody().endl();
		}
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t\t\tbreak;";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t\tdefault: ";
		if (_bContinue) {
			theCompilerEnvironment.getBody() << "CGRuntime::throwBNFExecutionError(";
			theCompilerEnvironment.getBody().writeString(toString());
			theCompilerEnvironment.getBody() << ");";
		} else {
			theCompilerEnvironment.getBody() << "_compilerClauseSuccess = false;";
		}
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t}";
		theCompilerEnvironment.getBody().endl();
	}
	if (_pVariableToAssign != 0L) {
		if (!_listOfConstants.empty() && !_bContinue) {
			theCompilerEnvironment.incrementIndentation();
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
			theCompilerEnvironment.getBody().endl();
		}
		compileCppAssignment(theCompilerEnvironment, iCursor);
		if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".pushItem(tcText);";
			theCompilerEnvironment.getBody().endl();
		} else {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".setValue(tcText);";
			theCompilerEnvironment.getBody().endl();
		}
		if (!_listOfConstants.empty() && !_bContinue) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
			theCompilerEnvironment.getBody().endl();
			theCompilerEnvironment.decrementIndentation();
		}
	}
	theCompilerEnvironment.incrementIndentation();
	if (!_listOfConstants.empty() && !_bContinue) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) {";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::setLocation(_compilerClauseLocation_" << iCursor << ");";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tif (_compilerClauseImplicitCopy_" << iCursor << " >= 0) CGRuntime::resizeOutputStream(_compilerClauseImplicitCopy_" << iCursor << ");";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "} else ";
	} else {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation();
	}
	theCompilerEnvironment.getBody() << "if (_compilerClauseImplicitCopy_" << iCursor << " >= 0) {";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tunsigned char cChar = (unsigned char) _compilerClauseChar_" << iCursor << ";";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::writeBinaryData((const char*) &cChar, 1);";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.decrementIndentation();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
	theCompilerEnvironment.getBody().endl();
}

void BNFReadChar::compileCppAssignment(CppCompilerEnvironment& theCompilerEnvironment, int iCursor) const {
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tchar tcText[] = {'\\0', '\\0'};";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\ttcText[0] = (char) _compilerClauseChar_" << iCursor << ";";
	theCompilerEnvironment.getBody().endl();
}

std::string BNFReadChar::toString() const {
	std::string sText = "#readChar" + DtaBNFScript::constantsToString(_listOfConstants) + DtaBNFScript::assignmentToString(_pVariableToAssign);
	if (_bContinue) sText = "#continue " + sText;
	return sText;
}
