/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif

#include "UtlException.h"
#include "ScpStream.h"
#include "CppCompilerEnvironment.h"
#include "CGRuntime.h"

#include "DtaScriptVariable.h"
#include "ExprScriptVariable.h"
#include "DtaBNFScript.h"
#include "BNFClause.h"
#include "BNFString.h"

using namespace CodeWorker;

BNFString::BNFString(DtaBNFScript* pBNFScript, GrfBlock* pParent, const std::string& sText, bool bContinue, bool bNoCase) : _pBNFScript(pBNFScript), GrfCommand(pParent), _pVariableToAssign(0L), _bContinue(bContinue), _iClauseReturnType(BNFClause::NO_RETURN_TYPE), _bNoCase(bNoCase) {
	if (_bNoCase) {
		_bNoCase = false;
		for (int i = 0; i < sText.size(); i++) {
			char a = sText[i];
			if ((a >= 'A') && (a <= 'Z')) {
				a += ' ';
				_bNoCase = true;
			} else if (!_bNoCase && (a >= 'a') && (a <= 'z')) _bNoCase = true;
			_sText += a;
		}
	} else {
		_sText = sText;
	}
}

BNFString::~BNFString() {
	delete _pVariableToAssign;
}

bool BNFString::isABNFCommand() const { return true; }

void BNFString::setVariableToAssign(ExprScriptVariable* pVariableToAssign, bool bConcat, BNFClause& theClause) {
	if (pVariableToAssign != 0L) {
		_pVariableToAssign = pVariableToAssign;
		std::string sVariableName = _pVariableToAssign->toString();
		if (sVariableName == theClause.getName()) _iClauseReturnType = theClause.getReturnType();
		_bConcatVariable = bConcat;
	}
}

SEQUENCE_INTERRUPTION_LIST BNFString::executeInternal(DtaScriptVariable& visibility) {
	bool bSuccess;
	int iLocation = CGRuntime::getLocation();
	int iImplicitCopyPosition = _pBNFScript->skipEmptyChars(visibility);
	if (_bNoCase) bSuccess = CGRuntime::readIfEqualToIgnoreCase(_sText);
	else bSuccess = CGRuntime::readIfEqualTo(_sText);
	if (!bSuccess) {
		if (_bContinue) CGRuntime::throwBNFExecutionError(toString());
		CGRuntime::setLocation(iLocation);
		if (iImplicitCopyPosition >= 0) CGRuntime::resizeOutputStream(iImplicitCopyPosition);
		return BREAK_INTERRUPTION;
	}
	if (_pVariableToAssign != 0L) {
		DtaScriptVariable* pVariable = visibility.getVariable(*_pVariableToAssign);
		if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
			pVariable->pushItem(_sText);
		} else {
			if (_bConcatVariable) pVariable->concatValue(_sText.c_str());
			else pVariable->setValue(_sText.c_str());
		}
	}
	if (iImplicitCopyPosition >= 0) _pBNFScript->writeBinaryData(_sText.c_str(), _sText.size());
	return NO_INTERRUPTION;
}

void BNFString::compileCpp(CppCompilerEnvironment& theCompilerEnvironment) const {
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "// " << toString();
	theCompilerEnvironment.getBody().endl();
	int iCursor = theCompilerEnvironment.newCursor();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseLocation_" << iCursor << " = CGRuntime::getLocation();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "int _compilerClauseImplicitCopy_" << iCursor << " = theEnvironment.skipEmptyChars();";
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "_compilerClauseSuccess = CGRuntime::readIfEqualTo";
	if (_bNoCase) theCompilerEnvironment.getBody() << "IgnoreCase";
	theCompilerEnvironment.getBody() << "(";
	theCompilerEnvironment.getBody().writeString(_sText);
	theCompilerEnvironment.getBody() << ");";
	theCompilerEnvironment.getBody().endl();
	if (_pVariableToAssign != 0L) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (_compilerClauseSuccess) {";
		theCompilerEnvironment.getBody().endl();
		if (_iClauseReturnType == BNFClause::LIST_RETURN_TYPE) {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".pushItem(";
			theCompilerEnvironment.getBody().writeString(_sText);
			theCompilerEnvironment.getBody() << ");";
			theCompilerEnvironment.getBody().endl();
		} else {
			theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\t";
			_pVariableToAssign->compileCppForBNFSet(theCompilerEnvironment);
			theCompilerEnvironment.getBody() << ".setValue(";
			theCompilerEnvironment.getBody().writeString(_sText);
			theCompilerEnvironment.getBody() << ");";
			theCompilerEnvironment.getBody().endl();
		}
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "}";
		theCompilerEnvironment.getBody().endl();
	}
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "if (!_compilerClauseSuccess) {";
	theCompilerEnvironment.getBody().endl();
	if (_bContinue) {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::throwBNFExecutionError(";
		theCompilerEnvironment.getBody().writeString(toString());
		theCompilerEnvironment.getBody() << ");";
	} else {
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tCGRuntime::setLocation(_compilerClauseLocation_" << iCursor << ");";
		theCompilerEnvironment.getBody().endl();
		theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "\tif (_compilerClauseImplicitCopy_" << iCursor << " >= 0) CGRuntime::resizeOutputStream(_compilerClauseImplicitCopy_" << iCursor << ");";
	}
	theCompilerEnvironment.getBody().endl();
	theCompilerEnvironment.getBody() << theCompilerEnvironment.getIndentation() << "} else if (_compilerClauseImplicitCopy_" << iCursor << " >= 0) CGRuntime::writeBinaryData(";
	theCompilerEnvironment.getBody().writeString(_sText);
	theCompilerEnvironment.getBody() << ", " << (int) _sText.size() << ");";
	theCompilerEnvironment.getBody().endl();
}

std::string BNFString::toString() const {
	std::string sText = "\"" + _sText + "\"" + DtaBNFScript::assignmentToString(_pVariableToAssign);
	if (_bContinue) sText = "#continue " + sText;
	return sText;
}
