/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning(disable : 4786)
#endif

#ifndef WIN32
#	include <cstdio> // for Debian/gcc 2.95.4
#endif

#include "ScpStream.h"
#include "UtlException.h"
#include "ExprScriptVariable.h"
#include "DtaScriptVariable.h"
#include "DtaScript.h"
#include "CppParsingTree.h"

using namespace CodeWorker;

CppParsingTree_expr::CppParsingTree_expr(const std::string& sExpression) {
	DtaScript theEmptyScript(0L);
	GrfBlock* pBlock = 0L;
	GrfBlock& myNullBlock = *pBlock;
	ScpStream myStream(sExpression);
	_pExpression = theEmptyScript.parseExpression(myNullBlock, myStream);
}

CppParsingTree_expr::~CppParsingTree_expr() {
	delete _pExpression;
}

CppParsingTree_varexpr::CppParsingTree_varexpr(const std::string& sVariableExpression) {
	DtaScript theEmptyScript(0L);
	GrfBlock* pBlock = 0L;
	GrfBlock& myNullBlock = *pBlock;
	ScpStream myStream(sVariableExpression);
	_pExpression = theEmptyScript.parseVariableExpression(myNullBlock, myStream);
}

ExprScriptVariable& CppParsingTree_varexpr::getVariableExpression() const {
	return *dynamic_cast<ExprScriptVariable*>(_pExpression);
}


CppParsingTree_var::CppParsingTree_var() : _pInternalNode(0L) {
}

CppParsingTree_var::CppParsingTree_var(const CppParsingTree_var& tree) {
	_pInternalNode = tree._pInternalNode;
}

CppParsingTree_var::CppParsingTree_var(DtaScriptVariable* pVariable) : _pInternalNode(pVariable) {
}

CppParsingTree_var::CppParsingTree_var(DtaScriptVariable& variable) : _pInternalNode(&variable) {
}

const char* CppParsingTree_var::getName() const {
	if (_pInternalNode == 0L) return 0L;
	return _pInternalNode->getName();
}

const std::string& CppParsingTree_var::getValue() const {
	static std::string sEmpty;
	if (_pInternalNode == 0L) return sEmpty;
	return _pInternalNode->getValue();
}

bool CppParsingTree_var::getBooleanValue() const {
	if (_pInternalNode == 0L) return 0;
	std::string sValue = _pInternalNode->getValue();
	return !sValue.empty();
}

int CppParsingTree_var::getIntValue() const {
	if (_pInternalNode == 0L) return 0;
	return atoi(_pInternalNode->getValue().c_str());
}

double CppParsingTree_var::getDoubleValue() const {
	if (_pInternalNode == 0L) return 0.0;
	return atof(_pInternalNode->getValue().c_str());
}

void CppParsingTree_var::setValue(const char* tcValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to set value \"" + std::string(tcValue) + "\" to an empty node");
	_pInternalNode->setValue(tcValue);
}

void CppParsingTree_var::setValue(const std::string& sValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to set value \"" + sValue + "\" to an empty node");
	_pInternalNode->setValue(sValue.c_str());
}

void CppParsingTree_var::setValue(double dValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to set a numeric to an empty node");
	_pInternalNode->setValue(dValue);
}

void CppParsingTree_var::setValue(bool bValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to set a boolean to an empty node");
	_pInternalNode->setValue((bValue ? "true" : ""));
}

void CppParsingTree_var::setValue(int iValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to set an integer to an empty node");
	_pInternalNode->setValue(iValue);
}

void CppParsingTree_var::setValue(const CppParsingTree_var& pValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to set a value to an empty node");
	if (pValue._pInternalNode == 0L) _pInternalNode->setValue("");
	_pInternalNode->setValue(pValue.getValue().c_str());
}

void CppParsingTree_var::setValue(const std::list<std::string>& listOfValues) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to set an array of strings to an empty node");
	_pInternalNode->setValue(listOfValues);
}

void CppParsingTree_var::concatenateValue(const std::string& sValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to concatenate value \"" + sValue + "\" to an empty node");
	std::string sConcatenation = _pInternalNode->getValue() + sValue;
	_pInternalNode->setValue(sConcatenation.c_str());
}

void CppParsingTree_var::concatenateValue(const CppParsingTree_var& pValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to concatenate a value to an empty node");
	std::string sConcatenation = _pInternalNode->getValue() + pValue.getValue();
	_pInternalNode->setValue(sConcatenation.c_str());
}

void CppParsingTree_var::setReference(CppParsingTree_var& reference) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to assign a reference to an empty node");
	if (reference._pInternalNode == 0L) throw UtlException("internal error: trying to assign an empty node as reference");
	_pInternalNode->setValue(reference._pInternalNode);
}

void CppParsingTree_var::setAll(const CppParsingTree_var& tree) {
	if (tree._pInternalNode == 0L) throw UtlException("internal error: trying to assign an empty subtree to a node");
	_pInternalNode->copyAll(*tree._pInternalNode);
}

void CppParsingTree_var::merge(const CppParsingTree_var& tree) {
	if (tree._pInternalNode == 0L) throw UtlException("internal error: trying to assign an empty subtree to a node");
	_pInternalNode->copyAll(*tree._pInternalNode, true);
}

void CppParsingTree_var::pushItem(const std::string& sValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to push a value to the array of an empty node");
	_pInternalNode->pushItem(sValue);
}

void CppParsingTree_var::addElements(const CppParsingTree_var& theValue) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to add elements to the array of an empty node");
	if (theValue._pInternalNode != 0L) _pInternalNode->addArrayElements(*(theValue._pInternalNode));
}

const std::list<DtaScriptVariable*>& CppParsingTree_var::getArray() const {
	static std::list<DtaScriptVariable*> emptyList;
	if (_pInternalNode == 0L) return emptyList;
	return _pInternalNode->getArray();
}

const std::map<std::string, DtaScriptVariable*>& CppParsingTree_var::getSortedArray() const {
	static std::map<std::string, DtaScriptVariable*> emptyMap;
	if (_pInternalNode == 0L) return emptyMap;
	return _pInternalNode->getSortedArray();
}

std::auto_ptr<std::vector<DtaScriptVariable*> > CppParsingTree_var::getSortedNoCaseArray() const {
	if (_pInternalNode == 0L) return std::auto_ptr<std::vector<DtaScriptVariable*> >(new std::vector<DtaScriptVariable*>());
	return _pInternalNode->getSortedNoCaseArray();
}

void CppParsingTree_var::clearNode() {
	if (_pInternalNode != 0L) _pInternalNode->clearContent();
}

CppParsingTree_var CppParsingTree_var::getNode(const std::string& sBranch) const {
	if (_pInternalNode == 0L) return CppParsingTree_var();
	DtaScriptVariable* pVariable = _pInternalNode->getNode(sBranch.c_str());
	return CppParsingTree_var(pVariable);
}

CppParsingTree_var CppParsingTree_var::getParentNode() const {
	if (_pInternalNode == 0L) return CppParsingTree_var();
	DtaScriptVariable* pVariable = _pInternalNode->getParent();
	return CppParsingTree_var(pVariable);
}

CppParsingTree_var CppParsingTree_var::getArrayNodeFromKey(const std::string& sKey) const {
	if (_pInternalNode == 0L) return CppParsingTree_var();
	DtaScriptVariable* pVariable = _pInternalNode->getArrayElement(sKey);
	return CppParsingTree_var(pVariable);
}

CppParsingTree_var CppParsingTree_var::getArrayNodeFromPosition(int iPosition) const {
	if (_pInternalNode == 0L) return CppParsingTree_var();
	DtaScriptVariable* pVariable = _pInternalNode->getArrayElement(iPosition);
	return CppParsingTree_var(pVariable);
}

CppParsingTree_var CppParsingTree_var::getFirstArrayNode() const {
	if (_pInternalNode == 0L) return CppParsingTree_var();
	DtaScriptVariable* pVariable = _pInternalNode->getArrayElement((int) 0);
	return CppParsingTree_var(pVariable);
}

CppParsingTree_var CppParsingTree_var::getLastArrayNode() const {
	if (_pInternalNode == 0L) return CppParsingTree_var();
	DtaScriptVariable* pVariable = _pInternalNode->getArrayElement(_pInternalNode->getArraySize() - 1);
	return CppParsingTree_var(pVariable);
}

CppParsingTree_var CppParsingTree_var::insertNode(const std::string& sBranch) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to insert '" + sBranch + "' key node into an empty tree");
	DtaScriptVariable* pVariable = _pInternalNode->insertNode(sBranch.c_str());
	return CppParsingTree_var(pVariable);
}

CppParsingTree_var CppParsingTree_var::insertArrayNodeFromKey(const std::string& sKey) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to insert '" + sKey + "' key node into an empty array");
	DtaScriptVariable* pVariable = _pInternalNode->addElement(sKey);
	return CppParsingTree_var(pVariable);
}

CppParsingTree_var CppParsingTree_var::insertArrayNodeFromKey(int iKey) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to insert key node into an empty array");
	char tcNumber[16];
	sprintf(tcNumber, "%d", iKey);
	DtaScriptVariable* pVariable = _pInternalNode->addElement(tcNumber);
	return CppParsingTree_var(pVariable);
}

CppParsingTree_var CppParsingTree_var::insertArrayNodeFromKey(const CppParsingTree_var& pKey) {
	if (_pInternalNode == 0L) throw UtlException("internal error: trying to insert key node into an empty tree");
	DtaScriptVariable* pVariable = _pInternalNode->addElement(pKey.getValue());
	return CppParsingTree_var(pVariable);
}


CppParsingTree_value::CppParsingTree_value() {
	_pInternalNode = new DtaScriptVariable;
}

CppParsingTree_value::CppParsingTree_value(const char* tcValue) {
	_pInternalNode = new DtaScriptVariable;
	_pInternalNode->setValue(tcValue);
}

CppParsingTree_value::CppParsingTree_value(const CppParsingTree_var& tree) {
	_pInternalNode = new DtaScriptVariable;
	setValue(tree);
}

CppParsingTree_value::CppParsingTree_value(const CppParsingTree_value& tree) {
	_pInternalNode = new DtaScriptVariable;
	setValue(tree);
}

CppParsingTree_value::CppParsingTree_value(bool bValue) {
	_pInternalNode = new DtaScriptVariable;
	_pInternalNode->setValue(bValue);
}

CppParsingTree_value::CppParsingTree_value(int iValue) {
	_pInternalNode = new DtaScriptVariable;
	_pInternalNode->setValue(iValue);
}

CppParsingTree_value::CppParsingTree_value(const std::string& sValue) {
	_pInternalNode = new DtaScriptVariable;
	_pInternalNode->setValue(sValue.c_str());
}

CppParsingTree_value::~CppParsingTree_value() {
	delete _pInternalNode;
}
