/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif

#include "ScpStream.h"
#include "UtlException.h"
#include "UtlTrace.h"

#include "DtaScript.h"
#include "DtaProject.h"

using namespace CodeWorker;

DtaProject* DtaProject::_pInstance = 0L;
std::list<DtaProject*> DtaProject::_stackOfProjects = std::list<DtaProject*>();

DtaProject::DtaProject() :	DtaScriptVariable(0L, "project"), _pScript(0L),
							_bOverwriteMode(true), _sCommentBegin("//"), _sCommentEnd("\n"),
							_dLastDelay(0.0), _pReadonlyHook(0L), _pWritefileHook(0L)
{
	if (_pInstance != 0L) {
		_stackOfProjects.push_back(_pInstance);
		copyCommon(*_pInstance);
	} else {
#ifdef WIN32
		_iTextMode = DOS_MODE;
#else
		_iTextMode = UNIX_MODE;
#endif
	}
	_pInstance = this;
}

DtaProject::~DtaProject() {
	for (std::map<std::string, DtaScriptVariable*>::const_iterator i = _mapOfGlobalVariables.begin(); i != _mapOfGlobalVariables.end(); i++) {
		delete i->second;
	}
	if (_stackOfProjects.empty()) {
		_pInstance = 0L;
	} else {
		_pInstance = _stackOfProjects.back();
		_stackOfProjects.pop_back();
		_pInstance->copyCommon(*this);
	}
}

DtaProject& DtaProject::createRootInstance() {
	if (!existInstance()) new DtaProject;
	return getInstance();
}

const std::string& DtaProject::getDefineTarget(const std::string& sTarget) const {
	std::map<std::string, std::string>::const_iterator i = _mapOfDefinedTargets.find(sTarget);
	if (i != _mapOfDefinedTargets.end()) return i->second;
	return ((DtaProject*) this)->_mapOfDefinedTargets[""];
}

void DtaProject::setDefineTarget(const std::string& sTarget, const std::string& sValue) {
	if (sTarget.empty()) throw UtlException("empty property '=\"" + sValue + "\"' is forbidden");
	_mapOfDefinedTargets[sTarget] = sValue;
}

DtaScriptVariable* DtaProject::getGlobalVariable(const std::string& sName) const {
	std::map<std::string, DtaScriptVariable*>::const_iterator cursor = _mapOfGlobalVariables.find(sName);
	if (cursor == _mapOfGlobalVariables.end()) return 0L;
	return cursor->second;
}

DtaScriptVariable* DtaProject::setGlobalVariable(const std::string& sName) {
	DtaScriptVariable* pGlobalVariable;
	std::map<std::string, DtaScriptVariable*>::const_iterator cursor = _mapOfGlobalVariables.find(sName);
	if (cursor != _mapOfGlobalVariables.end()) {
		pGlobalVariable = cursor->second;
		pGlobalVariable->clearContent();
	} else {
		pGlobalVariable = new DtaScriptVariable(0L, sName);
		_mapOfGlobalVariables[sName] = pGlobalVariable;
	}
	return pGlobalVariable;
}

void DtaProject::setTextMode(TEXT_MODE iTextMode) {
	_iTextMode = iTextMode;
	if (iTextMode == DOS_MODE) {
		if (strcmp(ScpStream::ENDL.c_str(), "\r\n") != 0) ScpStream::ENDL = "\r\n";
	} else {
		if (strcmp(ScpStream::ENDL.c_str(), "\n") != 0) ScpStream::ENDL = "\n";
	}
}

void DtaProject::copyCommon(const DtaProject& project) {
	_pScript = project._pScript;
	_sVersion = project._sVersion;
	_sMarkupKey = project._sMarkupKey;
	_sMarkupValue = project._sMarkupValue;
	_bOverwriteMode = project._bOverwriteMode;
	_sWorkingPath = project._sWorkingPath;
	_mapOfDefinedTargets = project._mapOfDefinedTargets;
	_sCommentBegin = project._sCommentBegin;
	_sCommentEnd = project._sCommentEnd;
	_sGenerationHeader = project._sGenerationHeader;
	_mapOfWarnings = project._mapOfWarnings;
	_dLastDelay = project._dLastDelay;
	_pReadonlyHook = project._pReadonlyHook;
	_pWritefileHook = project._pWritefileHook;
	_bSpeed = project._bSpeed;
	_iTextMode = project._iTextMode;
}

std::string DtaProject::getTraceStack(ScpStream& script) const {
	std::string sStack = UtlTrace::getTraceStack();
	const char* tcFilename = getScript()->getFilenamePtr();
	if (tcFilename == 0L) {
		sStack += "line ";
	} else {
		sStack += "\"";
		sStack += tcFilename;
		sStack += "\" line ";
	}
	char tcNumber[16];
	sprintf(tcNumber, "%d", script.getLineCount());
	sStack += tcNumber;
	return sStack;
}

void DtaProject::traceEngine() const {
	if (_pScript != 0L) {
		_pScript->traceEngine();
	}
}

void DtaProject::prepareExecution() {
	_mapOfWarnings = std::map<std::string, int>();
}

int DtaProject::addWarning(const std::string& sWarning) {
	int& iCounter = _mapOfWarnings[sWarning];
	iCounter++;
	return iCounter;
}

bool DtaProject::isScriptVersionOlderThan(int* tiVersion) const {
	return true;
}
