/* "CodeWorker":	a scripting language for parsing and generating text.

Copyright (C) 1996-1997, 1999-2002 Cdric Lemaire

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

To contact the author: codeworker@free.fr
*/

#ifdef WIN32
#pragma warning (disable : 4786)
#endif

#include "UtlException.h"
#include "ScpStream.h"
#include "DtaProject.h"
#include "CGRuntime.h"
#include "DtaProtectedAreasBag.h"

using namespace CodeWorker;

DtaProtectedAreasBag::~DtaProtectedAreasBag() {
	clearAll();
}

void DtaProtectedAreasBag::clearAll() {
	for (std::map<std::string, DtaProtectedArea*>::const_iterator i = _codes.begin(); i != _codes.end(); i++) {
		if (i->second != 0L) delete i->second;
	}
	_codes = std::map<std::string, DtaProtectedArea*>();
}

std::string DtaProtectedAreasBag::getProtection(const char* sProtection) const {
	std::map<std::string, DtaProtectedArea*>::const_iterator i = _codes.find(sProtection);
	if (i != _codes.end()) return i->second->getText();
	return "";
}

void DtaProtectedAreasBag::setProtection(const char* sProtection, const char* sContent) {
	std::map<std::string, DtaProtectedArea*>::const_iterator i = _codes.find(sProtection);
	if (i != _codes.end()) {
		if (i->second->isAlreadyGenerated()) throw UtlException(std::string("protected area called \"") + sProtection + "\" has already been generated");
		i->second->setText(sContent);
	} else {
		DtaProtectedArea* pArea = new DtaProtectedArea(sContent, "");
		_codes[sProtection] = pArea;
	}
}

bool DtaProtectedAreasBag::removeProtection(const char* sProtection) {
	std::map<std::string, DtaProtectedArea*>::iterator cursor = _codes.find(sProtection);
	if (cursor == _codes.end()) return false;
	if (cursor->second->isAlreadyGenerated()) return false;
	cursor->second->isAlreadyGenerated(true);
	return true;
}

std::list<std::string> DtaProtectedAreasBag::getProtectionKeys() const {
	std::list<std::string> listOfKeys;
	for (std::map<std::string, DtaProtectedArea*>::const_iterator i = _codes.begin(); i != _codes.end(); i++) {
		if (i->second != 0L) listOfKeys.push_back(i->first);
	}
	return listOfKeys;
}

std::list<std::string> DtaProtectedAreasBag::remainingProtectionKeys() const {
	std::list<std::string> listOfKeys;
	for (std::map<std::string, DtaProtectedArea*>::const_iterator i = _codes.begin(); i != _codes.end(); i++) {
		register DtaProtectedArea* pArea = i->second;
		if ((pArea != 0L) && !pArea->isAlreadyGenerated()) {
			listOfKeys.push_back(i->first);
		}
	}
	return listOfKeys;
}

const DtaProtectedArea& DtaProtectedAreasBag::registerNewProtection(const char* sProtection) {
	std::map<std::string, DtaProtectedArea*>::const_iterator i = _codes.find(sProtection);
	DtaProtectedArea* pArea;
	if (i == _codes.end()) {
		pArea = new DtaProtectedArea("");
		_codes[sProtection] = pArea;
	} else {
		pArea = i->second;
		if (pArea->isAlreadyGenerated()) throw UtlException(std::string("protected area called \"") + sProtection + "\" has already been generated");
	}
	pArea->isAlreadyGenerated(true);
	return *pArea;
}

void DtaProtectedAreasBag::loadProtectedCodes(const char* sFile) {
	std::string sCompleteFileName;
	ScpStream* pInputFile = ScpStream::openInputFileFromIncludePath(sFile, sCompleteFileName);
	if (pInputFile != 0L) {
		try {
			recoverProtectedCodes(*pInputFile);
		} catch(UtlException& e) {
			std::string sMessage = sFile;
			sMessage += ", ";
			sMessage += e.getMessage();
			pInputFile->close();
			delete pInputFile;
			throw UtlException(e.getTraceStack(), sMessage);
		}
		pInputFile->close();
		delete pInputFile;
	}
}

void DtaProtectedAreasBag::recoverProtectedCodes(ScpStream& inputFile) {
	if (!DtaProject::getInstance().getCommentBegin().empty()) {
		std::string sBeginning = DtaProject::getInstance().getCommentBegin() + "##protect";
		int iLocation = inputFile.getInputLocation();
		inputFile.setInputLocation(0);
		while (inputFile.findString(sBeginning.c_str())) recoverProtectedCode(inputFile, sBeginning);
		inputFile.setInputLocation(iLocation);
	}
}

void DtaProtectedAreasBag::recoverMarker(ScpStream& inputFile, const std::string& sMarkerKey) {
	if (!DtaProject::getInstance().getCommentBegin().empty() && inputFile.isEqualTo(DtaProject::getInstance().getCommentBegin() + "##begin##")) {
		std::string sStringMarkerKey = "\"" + CGRuntime::composeCLikeString(sMarkerKey) + "\"";
		if (!inputFile.isEqualTo(sStringMarkerKey)) throw UtlException(inputFile, "'##begin##' tag is expected to concern marker called '" + sStringMarkerKey + "'");
		if (!inputFile.findString(DtaProject::getInstance().getCommentEnd())) throw UtlException(inputFile, "'##begin##' tag is expected to end with an end of comment marker");
		int iBegin = inputFile.getInputLocation();
		if (!inputFile.findString(DtaProject::getInstance().getCommentBegin() + "##end##"))  throw UtlException(inputFile, "'##begin##' tag is expected to have an associated '##end##' tag");
		int iEnd = inputFile.getInputLocation() - (7 + DtaProject::getInstance().getCommentBegin().size());
		if (!inputFile.isEqualTo(sStringMarkerKey)) throw UtlException(inputFile, "'##end##' tag is expected to concern marker called " + sStringMarkerKey);
		if (!inputFile.findString(DtaProject::getInstance().getCommentEnd())) throw UtlException(inputFile, "'##end##' tag is expected to end with an end of comment marker");
		inputFile.setInputLocation(iBegin);
		std::string sBeginning = DtaProject::getInstance().getCommentBegin() + "##protect";
		while (inputFile.findString(sBeginning.c_str(), iEnd)) recoverProtectedCode(inputFile, sBeginning);
		inputFile.setInputLocation(iEnd + (7 + DtaProject::getInstance().getCommentBegin().size()));
		if (!inputFile.isEqualTo(sStringMarkerKey)) throw UtlException(inputFile, "'##end##' tag is expected to concern marker called " + sStringMarkerKey);
		while (!inputFile.isEqualTo(DtaProject::getInstance().getCommentEnd())) {
			int a = inputFile.readChar();
			if ((a < 0) || (a > (int) ' ')) throw UtlException(inputFile, "end of comment expected after '##markup##\"" + sStringMarkerKey + "\"'");
		}
	}
}

void DtaProtectedAreasBag::recoverProtectedCode(ScpStream& inputFile, const std::string& sBeginning) {
	std::string sDefine;
	inputFile.skipEmpty();
	if (inputFile.readIdentifier(sDefine)) inputFile.skipEmpty();
	if (!inputFile.isEqualTo("##")) throw UtlException(inputFile, "wrong syntax for announcing protected area, '##' expected");
	std::string sText;
	if (!inputFile.readString(sText)) throw UtlException(inputFile, "wrong syntax for announcing protected area");
	if (_codes.find(sText) != _codes.end()) throw UtlException(inputFile, "a protected area called \"" + sText + "\" already exists");
	if (!inputFile.findString(DtaProject::getInstance().getCommentEnd().c_str())) throw UtlException(inputFile, "end of comment expected after announcing protected area \"" + sText + "\"");
	int iBeginCode = inputFile.getInputLocation();
	std::string sEndText;
	if (!inputFile.findString(sBeginning.c_str())) throw UtlException(inputFile, "end of protected area \"" + sText + "\" expected");
	std::string sCode = getCode(inputFile, iBeginCode);
	std::string sEndDefine;
	inputFile.skipEmpty();
	if (inputFile.readIdentifier(sEndDefine)) inputFile.skipEmpty();
	if (!inputFile.isEqualTo("##")) throw UtlException(inputFile, "wrong syntax for announcing protected area \"" + sText + "\", '##' expected");
	if (sDefine != sEndDefine) throw UtlException(inputFile, "wrong syntax for announcing protected area \"" + sText + "\", defines are different");
	if (!inputFile.readString(sEndText)) throw UtlException(inputFile, "wrong syntax for finishing protected area \"" + sText + "\"");
	if (!inputFile.findString(DtaProject::getInstance().getCommentEnd().c_str())) throw UtlException(inputFile, "end of comment expected after finishing protected area \"" + sText + "\"");
	if (sText != sEndText) throw UtlException(inputFile, "marker \"" + sText + "\" of beginning of area doesn't match with marker of end of area (\"" + sEndText + "\")");
	if (!sCode.empty()) {
		DtaProtectedArea* pArea = new DtaProtectedArea(sCode, sDefine);
		_codes[sText] = pArea;
	}
}

std::string DtaProtectedAreasBag::getCode(ScpStream& inputFile, int iBeginCode) {
	std::string sCode;
	int iPosition = inputFile.getInputLocation();
	int iEndCode = iPosition;
	while (iEndCode > iBeginCode) {
		iEndCode--;
		inputFile.setInputLocation(iEndCode);
		if (inputFile.readChar() == (int) '\n') {
			iEndCode++;
			break;
		}
	}
	inputFile.setInputLocation(iBeginCode);
	for (int iCurrent = iBeginCode; iCurrent < iEndCode; iCurrent++) sCode += (char) inputFile.readChar();
	inputFile.setInputLocation(iPosition);
	return sCode;
}

